import functools
import itertools
from functools import singledispatchmethod
from typing import Dict, Any, List, Union

import jax
from flax import struct
from jax import numpy as jnp

from semistaticsim.groundtruth.constants import AFFORDANCE_NAMES
from semistaticsim.groundtruth.spoof_vmap import spoof_vmap
from semistaticsim.rendering.utils.dict2xyztup import dict2xyztuple, xyztuple2dict


def trim_spawn_coordinates(coords):
    # 1. Define how much to trim (5% from each side = 10% total removed)
    margin_percent = 0.1

    # Calculate bounds only for X and Z (indices 0 and 2)
    # We calculate full bounds, but only apply the mask to X and Z
    x_min, y_min, z_min = coords.min(axis=0)
    x_max, y_max, z_max = coords.max(axis=0)

    x_margin = (x_max - x_min) * margin_percent
    z_margin = (z_max - z_min) * margin_percent

    # Mask: Only filter based on X and Z borders
    mask = (
            (coords[:, 0] > x_min + x_margin) & (coords[:, 0] < x_max - x_margin) &
            (coords[:, 2] > z_min + z_margin) & (coords[:, 2] < z_max - z_margin)
    )

    inner_coords = coords[mask]
    return inner_coords

def oobb_to_aabb(oobb_cornerPoints: Union[tuple, list, jnp.array]):
    if isinstance(oobb_cornerPoints, jnp.ndarray):
        pass
    else:
        oobb_cornerPoints = jnp.array(oobb_cornerPoints)

    mins = jnp.min(oobb_cornerPoints, axis=0)  # (3,)
    maxs = jnp.max(oobb_cornerPoints, axis=0)  # (3,)
    bounds = jnp.stack([mins, maxs])
    rotated_aabb = jnp.array(
        [bounds[choice, jnp.arange(3)] for choice in itertools.product([0, 1], repeat=3)],
        dtype=jnp.float32,
    )

    # fifth, compute new AABB size (doesnt matter that it got translated)
    sizes = maxs - mins

    # sixth, compute new AABB center
    new_center = rotated_aabb.mean(axis=0)

    return {"center": xyztuple2dict(new_center), "size": xyztuple2dict(sizes), "cornerPoints": oobb_cornerPoints}



@struct.dataclass
class ProcThorObject:
    # id: str = struct.field(pytree_node=False)  # str bytes
    affordances: jnp.ndarray
    position: jnp.ndarray
    rotation: jnp.ndarray
    aabb_center: jnp.ndarray
    aabb_size: jnp.ndarray
    aabb_cornerPoints: jnp.ndarray
    oobb_cornerPoints: jnp.ndarray
    spawn_coordinates_above_receptacle: jnp.ndarray

    @property
    def is_pickupable(self):
        # FIXME: if self.affordances is 2D array, this will break
        return self.affordances[AFFORDANCE_NAMES.index("pickupable")]

    @property
    def is_receptacle(self):
        # FIXME: if self.affordances is 2D array, this will break
        return self.affordances[AFFORDANCE_NAMES.index("receptacle")]

    def take(self, i: int):
        values, tree_def = jax.tree.flatten(self)
        values = [jnp.take(v, jnp.astype(i, int), axis=0) for v in values]
        return jax.tree.unflatten(tree_def, values)

    @singledispatchmethod
    @classmethod
    def create(cls, arg):
        raise NotImplementedError()

    @classmethod
    def create_from_spawncoords_only(cls, spawn_coords):
        return cls(None,None,None,None,None,None,None,spawn_coords)

    @create.register(dict)
    @classmethod
    def _(cls, dict: Dict[str, Any]):
        # id = dict["id"]
        position = dict["position"]
        rotation = dict["rotation"]
        aabb = dict["aabb"]
        oobb = dict["oobb"]

        if oobb is None:
            oobb = aabb["cornerPoints"]

        affordances = []
        for name in AFFORDANCE_NAMES:
            val = dict["affordances"][name]
            affordances.append(bool(val))
        affordances = jnp.array(affordances)

        position = jnp.array(dict2xyztuple(position))
        rotation = jnp.array(dict2xyztuple(rotation))

        spawn_coordinates_above_receptacle = jnp.array(dict["spawn_coordinates_above_receptacle"])

        return cls(
            affordances,
            position,
            rotation,
            jnp.array(dict2xyztuple(aabb["center"])),
            jnp.array(dict2xyztuple(aabb["size"])),
            jnp.array(aabb["cornerPoints"]),
            jnp.array(oobb["cornerPoints"]),
            spawn_coordinates_above_receptacle
        )

    @create.register(list)
    @classmethod
    def _(cls, list: List[Dict]):
        return spoof_vmap([ProcThorObject.create(dico) for dico in list])

    # @jax.jit
    def sample_from_surface(self, key, num_samples=1):
        return jax.random.choice(key, self.spawn_coordinates_above_receptacle, shape=(num_samples,), replace=False)

    #@jax.jit
    def make_potential_field_OLD(self, current_positions):
        if len(current_positions) == 0:
            return jnp.ones((len(self.spawn_coordinates_above_receptacle),)) / len(self.spawn_coordinates_above_receptacle)

        # potential field input is a vector of positions. The ones at -jnp.inf are to be ignored. The others are to repulse the spawn coords
        def foreach_spawn(pots, spawn_coord):
            def foreach_pot(s, p):
                return jnp.linalg.norm(s - p)

            return jax.vmap(functools.partial(foreach_pot, spawn_coord))(pots)

        dists = jax.vmap(functools.partial(foreach_spawn, current_positions))(self.spawn_coordinates_above_receptacle)
        dists = jnp.nan_to_num(dists, posinf=0).sum(axis=1)

        remove_0_mask = dists == 0
        temp = dists.max() * remove_0_mask + dists * (1 - remove_0_mask)
        dists = dists.clip(
            temp.min(), max=jnp.inf
        )  # removes already selected points (they were 0's which made for a very spiky field)

        weights = dists * 10 ** 2  # makes it VERY likely to select the furthest point
        weights = weights / weights.sum()
        return weights

    import jax
    import functools

    def make_potential_field(self, current_positions, object_radius=0.1):
        """
        Args:
            current_positions: (M, 3) array of existing object positions.
            object_radius: Float radius to define a "hard collision".
        """
        num_spawns = len(self.spawn_coordinates_above_receptacle)

        # 1. Handle empty scene
        if len(current_positions) == 0:
            return jnp.ones((num_spawns,)) / num_spawns

        # 2. Calculate Distance Matrix (N_spawn x M_current)
        # We create a mask for valid objects (ignoring those at -inf)
        # Assuming objects at -inf are padding/invalid
        valid_obj_mask = jnp.all(jnp.isfinite(current_positions), axis=1)  # Shape (M,)

        def dist_fn(spawn, curr_pos):
            return jnp.linalg.norm(spawn - curr_pos)

        # Vectorize over spawns, then over current_positions
        # resulting shape: (N_spawns, M_current_positions)
        all_dists = jax.vmap(
            lambda s: jax.vmap(lambda c: dist_fn(s, c))(current_positions)
        )(self.spawn_coordinates_above_receptacle)

        # 3. Apply Safety Mask (The "Min" Step)
        # If an object is invalid (-inf), we set its distance to infinity so it doesn't affect the min()
        safe_dists = jnp.where(valid_obj_mask, all_dists, jnp.inf)

        # Find the distance to the *nearest* object for each spawn point
        min_dists = jnp.min(safe_dists, axis=1)  # Shape (N_spawns,)

        # 4. Filter Collisions (Hard Constraint)
        # Any point closer than object_radius is strictly invalid (weight 0)
        valid_spawn_mask = min_dists > object_radius

        # 5. Convert to Probabilities (Softmax-like scaling)
        # We want larger distances -> Higher probability
        # We use a temperature factor to sharpen the peak (greedy selection)

        # Mask out invalid spots by setting them to -inf before softmax
        # or simply zeroing them out after linear scaling. Let's do linear scaling for stability.

        weights = min_dists * valid_spawn_mask

        # Power scaling makes the field "spiky" (favoring the absolute furthest point)
        # increasing the power (e.g., **4) makes it more greedy.
        weights = weights ** 4

        # Normalize
        weight_sum = weights.sum()

        # Safety check: if all points are invalid (weight_sum is 0), return uniform distribution
        return jax.lax.cond(
            weight_sum > 0,
            lambda: weights / weight_sum,
            lambda: jnp.ones((num_spawns,)) / num_spawns
        )

    @functools.partial(jax.jit, static_argnums=(3,))
    def sample_from_surface_with_potential_field(self, key, potential_field, num_samples=1):
        # weights = self.make_potential_field(potential_field)
        ret = jax.random.choice(
            key, self.spawn_coordinates_above_receptacle, shape=(num_samples,), replace=False, p=potential_field
        )
        return ret

    @property
    def height(self):
        return jnp.atleast_1d(self.aabb_size[1]).squeeze()

    @property
    def width(self):
        return jnp.atleast_1d(self.aabb_size[0]).squeeze()

    @property
    def depth(self):
        return jnp.atleast_1d(self.aabb_size[2]).squeeze()

    def get_aabb(self):
        return {"center": xyztuple2dict(self.aabb_center), "size": xyztuple2dict(self.aabb_size),
         "cornerPoints": self.aabb_cornerPoints.tolist()}

    def get_oobb(self):
        return {"cornerPoints": self.oobb_cornerPoints.tolist()}

    @jax.jit
    def resolve_bboxes(self, new_bottom_pos, new_y_axis_rotation):
        if len(new_bottom_pos.squeeze().shape) == 2:
            return jax.vmap(functools.partial(ProcThorObject.resolve_bboxes, self))(new_bottom_pos, new_y_axis_rotation)

        # first, center the object
        current_oobb_center = jnp.mean(self.oobb_cornerPoints, axis=0)
        centered_oobb = self.oobb_cornerPoints - current_oobb_center

        # second, rotate the object around the origin
        cos_t, sin_t = jnp.cos(new_y_axis_rotation), jnp.sin(new_y_axis_rotation)
        rot_mat = jnp.array([[cos_t, 0.0, sin_t], [0.0, 1.0, 0.0], [-sin_t, 0.0, cos_t]], dtype=jnp.float32)  # (3, 3)
        rotated_oobb = centered_oobb @ rot_mat.T  # (8, 3)

        # fourth, translate by bottom_pos (NOT by center!)
        rotated_oobb = rotated_oobb + new_bottom_pos

        aabb = oobb_to_aabb(rotated_oobb)
        aabb["center"] = jnp.array(dict2xyztuple(aabb["center"]))
        aabb["size"] = jnp.array(dict2xyztuple(aabb["size"]))
        aabb["cornerPoints"] = jnp.array(aabb["cornerPoints"])
        oobb = {"cornerPoints": rotated_oobb}
        return aabb, oobb



import jax.numpy as jnp
from jax import jit
import jax

def _bool_ifelse_elementwise(cond, iftrue, iffalse):
    return iftrue * cond + iffalse * (1-cond)

@jax.jit
def bool_ifelse(cond, iftrue, iffalse):
    cond = jnp.atleast_1d(cond)
    iftrue = jnp.atleast_1d(iftrue)
    iffalse = jnp.atleast_1d(iffalse)

    MAIN_SHAPE = cond.shape[0]

    if len(iffalse.shape) == 0:
        iffalse = jnp.ones_like(cond) * iffalse

    if len(iftrue.shape) == 0:
        iftrue = jnp.ones_like(cond) * iftrue

    if iffalse.shape[0] != MAIN_SHAPE:
        iffalse = jnp.repeat(iffalse[None], MAIN_SHAPE, axis=0)
    if iftrue.shape[0] != MAIN_SHAPE:
        iftrue = jnp.repeat(iftrue[None], MAIN_SHAPE, axis=0)

    cond = cond.astype(int)

    return jax.vmap(_bool_ifelse_elementwise)(cond, iftrue, iffalse)


def isOnTop(pos1, size1, pos2, size2, tol_overlap=0.1, tol_dist=0.2):
    """Returns True if obj1 is on top of obj2 within given tolerances.

    Args:
        pos1: (x,y,z) position of obj1 (center coordinates)
        size1: (dx,dy,dz) size of obj1 (full widths/extents)
        pos2: (x,y,z) position of obj2 (center coordinates)
        size2: (dx,dy,dz) size of obj2 (full widths/extents)
        tol_overlap: minimum required XY overlap ratio (of obj1's area)
        tol_dist: maximum allowed vertical gap between objects

    Returns:
        True if obj1 is on top of obj2 according to tolerances
    """
    # Compute XY overlap as the intersection over the area of obj1
    overlap_x = jnp.maximum(0, jnp.minimum(pos1[0] + size1[0] / 2, pos2[0] + size2[0] / 2) -
                            jnp.maximum(pos1[0] - size1[0] / 2, pos2[0] - size2[0] / 2))
    overlap_y = jnp.maximum(0, jnp.minimum(pos1[1] + size1[1] / 2, pos2[1] + size2[1] / 2) -
                            jnp.maximum(pos1[1] - size1[1] / 2, pos2[1] - size2[1] / 2))
    overlap_area = overlap_x * overlap_y
    obj1_area = size1[0] * size1[1]
    #obj2_area = size2[0] * size2[1]

    # Use minimum of both areas for more robust overlap checking
    xy_overlap_ratio = overlap_area / obj1_area # jnp.maximum(1e-6, jnp.minimum(obj1_area, obj2_area))

    # Check vertical relationship
    top_z_obj2 = pos2[2] + size2[2] / 2
    bottom_z_obj1 = pos1[2] - size1[2] / 2
    z_close = (bottom_z_obj1 >= top_z_obj2 - tol_dist) & (bottom_z_obj1 <= top_z_obj2 + tol_dist)

    # Check if obj1 is above obj2 (using centers is not ideal, but better than nothing)
    is_above = bottom_z_obj1 > pos2[2] - size2[2] / 2  # obj1's bottom is above obj2's middle

    return (xy_overlap_ratio >= tol_overlap) & z_close & is_above

def isInside(pos1, size1, pos2, size2, tol=1e-6):
    """
    Strict containment check if obj1 is completely inside obj2.

    Args:
        pos1: (x,y,z) position of obj1 (center coordinates)
        size1: (dx,dy,dz) size of obj1 (full widths/extents)
        pos2: (x,y,z) position of obj2 (center coordinates)
        size2: (dx,dy,dz) size of obj2 (full widths/extents)
        tol: tolerance for boundary comparisons

    Returns:
        True if obj1 is completely inside obj2 (with tolerance), False otherwise
    """
    # Calculate the min and max bounds for each object
    obj1_min = pos1 - size1 / 2
    obj1_max = pos1 + size1 / 2
    obj2_min = pos2 - size2 / 2
    obj2_max = pos2 + size2 / 2

    # Check if all of obj1's bounds are within obj2's bounds with tolerance
    return (jnp.all(obj1_min >= obj2_min - tol) & jnp.all(obj1_max <= obj2_max + tol))


def isBeside(pos1, size1, pos2, size2, tol_dist=0.5):
    """Check if objects overlap in one axis and are close in the other."""
    same_surface = jnp.abs((pos1[2] - size1[2] / 2) - (pos2[2] - size2[2] / 2)) <= (tol_dist / 4)

    # Overlap in X and proximity in Y (or vice versa)
    overlap_x = (jnp.abs(pos1[0] - pos2[0]) <= (size1[0] + size2[0]) / 2)
    proximity_y = (jnp.abs(pos1[1] - pos2[1]) <= (size1[1] + size2[1]) / 2 + tol_dist)

    # Overlap in Y and proximity in X (or vice versa)
    overlap_y = (jnp.abs(pos1[1] - pos2[1]) <= (size1[1] + size2[1]) / 2)
    proximity_x = (jnp.abs(pos1[0] - pos2[0]) <= (size1[0] + size2[0]) / 2 + tol_dist)

    return same_surface & ((overlap_x & proximity_y) | (overlap_y & proximity_x))


def distance(pos1, size1, pos2, size2, tol_overlap=None, tol_dist=None):
    """Returns the Euclidean distance between obj1 and obj2."""
    return jnp.linalg.norm(pos1 - pos2)

def bbox_dist(pos1, size1, pos2, size2, tol_overlap=None, tol_dist=None):
    """
    Compute the minimal distance between two axis-aligned bounding boxes.

    Args:
        pos1: Center position of first box (shape: [d])
        size1: Size of first box (shape: [d])
        pos2: Center position of second box (shape: [d])
        size2: Size of second box (shape: [d])
        tol_overlap: If boxes overlap by more than this, return 0 (optional)
        tol_dist: If boxes are separated by more than this, return this value (optional)

    Returns:
        The minimal distance between the boxes, with optional thresholds applied
    """
    # Calculate the distance between centers
    delta = jnp.abs(pos1 - pos2)

    size1 = size1 / 2
    size2 = size2 / 2

    # Calculate the sum of half-extents
    sum_half_extents = size1 + size2

    # Compute separation in each dimension
    separation = delta - sum_half_extents

    # The minimal distance is the maximum separation across dimensions
    # (negative values indicate overlap)
    max_separation = jnp.max(separation)

    # If boxes overlap in all dimensions, the minimal distance is negative
    # (we take the maximum separation as the penetration depth)
    min_dist = jnp.where(max_separation < 0,
                         max_separation,  # penetration depth (negative)
                         jnp.linalg.norm(jnp.maximum(separation, 0)))  # positive distance

    # Apply tolerance for overlap if provided
    if tol_overlap is not None:
        min_dist = jnp.where(min_dist < -tol_overlap, 0.0, min_dist)

    # Apply tolerance for distance if provided
    if tol_dist is not None:
        min_dist = jnp.where(min_dist > tol_dist, tol_dist, min_dist)

    return min_dist


import functools
import json
from functools import partial

import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import networkx as nx
import shapely
from flax import struct
from shapely.geometry import Polygon
from shapely.ops import triangulate


@partial(jax.jit, static_argnums=(1,))
def split_key(key, num_keys):
    key, *rng = jax.random.split(key, num_keys + 1)
    rng = jnp.reshape(jnp.stack(rng), (num_keys, 2))
    return key, rng

def point_to_shapely(point):
    if len(point) == 3:
        def cast(p):
            return shapely.Point(p[0], p[2])
    else:
        def cast(p):
            return shapely.Point(p[0], p[1])
    return cast(point)



@struct.dataclass
class JaxPolygon:
    coords_polygon: list[tuple[float, float]]
    triangles: jnp.ndarray
    triangle_areas: jnp.ndarray

    @property
    def shapely_polygon(self):
        return shapely.Polygon(self.coords_polygon)

    @classmethod
    def create(cls, coords):
        if isinstance(coords[0], dict):
            coords = list(map(lambda x: (x["x"], x["z"]), coords))
        self = JaxPolygon(coords, None, None)

        polygon = self.shapely_polygon
        def polygon_to_triangles(polygon):
            """
            Triangulate a shapely polygon and return list of triangles.
            """
            tris = shapely.ops.triangulate(polygon)

            ret = []
            for tri in tris:
                centroid = tri.centroid
                if not polygon.contains(centroid):
                    continue
                ret.append(list(tri.exterior.coords)[:-1])

            return ret
        tris = jnp.array(polygon_to_triangles(polygon))

        def triangle_area(tri):
            x1, y1 = tri[0]
            x2, y2 = tri[1]
            x3, y3 = tri[2]
            return 0.5 * jnp.abs((x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1))
        triangle_areas = jax.vmap(triangle_area)(tris)

        return self.replace(triangles=tris, triangle_areas=triangle_areas)

    @property
    def num_triangles(self):
        return self.triangles.shape[0]

    @property
    def bbox(self):
        coords = self.coords_polygon

        min_x, min_y = jnp.inf, jnp.inf
        max_x, max_y = -jnp.inf, -jnp.inf

        for coord in coords:
            min_x = min(min_x, coord[0])
            min_y = min(min_y, coord[1])
            max_x = max(max_x, coord[0])
            max_y = max(max_y, coord[1])

        return min_x, min_y, max_x, max_y

    def containment_mask(self, points):
        return [self.shapely_floor_polygon.contains(point_to_shapely(point)) for point in points]

    def points_distance_to_polygon_edge(self, points):
        polygon = self.shapely_floor_polygon
        return [point_to_shapely(point).distance(polygon.boundary) for point in points]

    def points_distance_to_polygon_edge_mask(self, points, mindist):
        dists = self.points_distance_to_polygon_edge(points)
        return [d > mindist for d in dists]

    @property
    def floor_polygon_coords(self):
        return list(map(lambda x: (x["x"], x["z"]), self.coords_polygon))

    @property
    def shapely_floor_polygon(self):
        return Polygon(self.floor_polygon_coords)

    @property
    def shapely_triangles(self):
        ret = []
        for tri in self.triangles:
            ret.append(Polygon(tri))
        return ret

    def plot_self(self, samples=None, plot_triangles=False, reachability_graph=None, show=False):
        plt.plot(*self.shapely_floor_polygon.exterior.xy)
        if plot_triangles:
            for tri in self.shapely_triangles:
                x, y = tri.exterior.xy
                plt.fill(x, y, alpha=0.2)

        if reachability_graph is not None:

            def draw_grid_graph_2d(G, path=None, node_size=100, node_color='lightblue', edge_color='gray',
                                   show_labels=False):
                pos = {node: (node[2], node[0]) for node in G.nodes}  # node positions are their coordinates

                nx.draw(G, pos,
                        node_size=node_size,
                        node_color=node_color,
                        edge_color=edge_color,
                        with_labels=show_labels,
                        font_size=8)

                # If a path is provided, draw it in red
                if path:
                    # Draw path edges first (so nodes will be on top)
                    path_edges = list(zip(path[:-1], path[1:]))
                    nx.draw_networkx_edges(G, pos,
                                           edgelist=path_edges,
                                           edge_color='red',
                                           width=2)
                    # Draw path nodes
                    nx.draw_networkx_nodes(G, pos,
                                           nodelist=path,
                                           node_size=node_size,
                                           node_color='red')

            draw_grid_graph_2d(reachability_graph)

        if samples is not None:
            plt.scatter(samples[:,0], samples[:,1], c="orange", zorder=10)

        if show:
            plt.show()

    @functools.partial(jax.jit, static_argnums=2)
    def sample_from_floor(self, key, num_samples=1):
        if num_samples > 1:
            _, keys = split_key(key, num_samples)
            return jax.vmap(self.sample_from_floor)(keys).squeeze()

        probs = self.triangle_areas / self.triangle_areas.sum()

        key_tri, key_uv = jax.random.split(key)
        tri_indices = jax.random.choice(key_tri, self.triangles.shape[0], (1,), p=probs)
        chosen_tris = self.triangles[tri_indices]

        uvs = jax.random.uniform(key_uv, (1, 2))
        u = jnp.sqrt(uvs[:, 0])
        v = uvs[:, 1]
        w0 = 1 - u
        w1 = u * (1 - v)
        w2 = u * v
        weights = jnp.stack([w0, w1, w2], axis=-1)
        weights_expanded = weights[:, :, jnp.newaxis]

        sampled_points = jnp.sum(chosen_tris * weights_expanded, axis=1)
        return sampled_points  # Returns [1, 2] array

if __name__ == "__main__":
    curdir = "/".join(__file__.split("/")[:-1])
    # write_object_affordances_for_procthor(curdir)

    with open(curdir + "/object_data/test/0/objects.json", "r") as f:
        objects = json.load(f)
    x = objects[0]["oobb"]



    JaxPolygon.create(objects[0])


import json
import os
import random
import sys
import time
from typing import Dict
from tqdm import trange
from ai2thor.controller import Controller
from ai2thor.platform import CloudRendering
from traitlets import Any

from semistaticsim.datawrangling.sssd import load_sssd, path_2_parts
from semistaticsim.groundtruth.utils import seed_everything
from semistaticsim.rendering.floor import ReachabilityGraph

from semistaticsim.rendering.simulation.skill_simulator import Simulator

DEFAULT_ROBOT_HEIGHT = 1.0
DEFAULT_ROBOT_ROT = 90


def _get_self_install_dir():
    filepath = __file__
    return "/".join(filepath.split("/")[:-2])


def _get_ai2thor_install_dir():
    return _get_self_install_dir() + "/ai2thor"


def _get_ai2thor_install_build_dir():
    return _get_ai2thor_install_dir() + "/unity/builds"


def _get_ai2thorbuilds_dir(which="fixed"):
    return _get_self_install_dir() + f"/ai2thorbuilds/{which}"


def get_hippo_controller(scene, target_dir=None, **kwargs):
    # if isinstance(scene, str):
    #    if scene.endswith(".json"):
    #        with open(scene, "r") as f:
    #            scene = json.load(f)
    #    elif "procthor" in scene.lower():
    #        num = int(scene.replace("procthor", "")) #
    #        from procthorprocessing.procthor_utils import get_procthor10k
    #        scene = get_procthor10k()[num]

    # try:
    #    os.unlink( _get_ai2thor_install_build_dir())
    # except Exception as e:
    #    pass
    # assert not os.path.exists(_get_ai2thor_install_build_dir())
    # os.symlink(_get_ai2thorbuilds_dir(), _get_ai2thor_install_build_dir(), target_is_directory=True)

    cuda_visible_devices = list(
        map(
            int,
            filter(
                lambda y: y.isdigit(),
                map(
                    lambda x: x.strip(),
                    os.environ.get("CUDA_VISIBLE_DEVICES", "").split(","),
                ),
            ),
        )
    )

    print("\n\n~~~AI2THOR GPU NOTICE~~~\n\n")
    if len(cuda_visible_devices) > 0:
        print("AI2Thor controller will use GPU(s):", cuda_visible_devices)
        print("If this occurs on a cluster without a monitor, you need to set up a virtual display.")
        print("You can also diagnose this when facing a `vulkaninfo` error. Remove access to the GPU!")
    else:
        print("AI2Thor controller will use CPU only.")
    print("\n\n~~~~~~~~~~~~~~~~~~~~~~~\n\n")

    # Setup Platform Config
    if sys.platform != "darwin":
        ai2_platform = CloudRendering
    else:
        ai2_platform = None

    # Force controller to initially when random wait occurs
    controller = None
    max_retries = 5
    for attempt in range(max_retries):
        try:
            def find_free_port_in_range(start=1024, end=65535):
                import socket

                sock = socket.socket()
                sock.bind(("", 0))
                return sock.getsockname()[-1]

            ai2thor_port = find_free_port_in_range()
            print("Free port found for AI2Thor controller:", ai2thor_port)

            # os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = "/home/mila/c/charlie.gauthier/Holodeck/venv/lib/python3.10/site-packages/cv2/qt/plugins"
            # os.environ["QT_QPA_PLATFORM"] = "/home/mila/c/charlie.gauthier/Holodeck/venv/lib/python3.10/site-packages/cv2/qt/fonts"  # use offscreen rendering to avoid X11 issues
            # os.environ["KMP_DUPLICATE_LIB_OK"] = "True"
            # os.environ["TCM_ENABLE"] = "1"
            # os.environ["KMP_INIT_AT_FORK"] = "FALSE"
            # os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"  # suppress TensorFlow warnings
            # os.environ["TOKENIZERS_PARALLELISM"] = "false"
            os.environ["XDF_RUNTIME_DIR"] = "/tmp"  # fix for X11 issues in some environments

            # if "ENVIRONMENT" in os.environ:
            #    os.environ.pop("ENVIRONMENT")
            # print("Unsetting slurm variables...")
            # yay jank :)
            # UNSET_SLURM = []
            # for k, v in os.environ.items():
            #    if k.startswith("SLURM"):
            #        os.environ.pop(k)
            #        #UNSET_SLURM.append(f"{k}")
            # if len(UNSET_SLURM) > 0:
            #    UNSET_SLURM = "env -u " + " -u ".join(UNSET_SLURM)
            # else:
            #    UNSET_SLURM = ""
            # print("WILL UNSET SLURM ENV VARS:", UNSET_SLURM)

            print(
                "Local executable path for AI2Thor controller:",
                f"{_get_ai2thor_install_build_dir()}/thor-Linux64-local/thor-Linux64-local",
            )

            print(
                "Attempting to create AI2Thor controller... this might time out. If so, you need to allocate more cpu or ram or figure something else out."
            )
            # Random start delay to prevent simultaneous port binding (parallel runs)
            if attempt > 0: 
                time.sleep(random.uniform(0.5, 1.5))

            print(f"Attempt {attempt + 1}/{max_retries}: init AI2Thor on port {ai2thor_port}...")

            controller = Controller(
                port=ai2thor_port,
                platform=ai2_platform,
                agentMode="default",
                makeAgentsVisible=False,
                scene=scene,
                server_start_timeout=30,
                **kwargs,
            )
            print(f"Success: AI2Thor Controller active on port {ai2thor_port}.")
            return controller
        
        except Exception as e:
            print(f"Warning: Controller init failed on port {ai2thor_port}. Error: {e}")

            # Check if this was the last attempt
            if attempt == max_retries - 1:
                print("CRITICAL: Failed to initialize AI2-THOR after max retries.")
                print("Exiting...")
                os._exit(1)

            # Exponential backoff: Wait 2s, 4s, 8s...
            wait_time = (2 ** attempt) + random.uniform(0, 1)
            print(f"Retrying in {wait_time:.2f}s...")
            time.sleep(wait_time)


def get_list_of_objects(scene):
    with open(scene, "r") as f:
        scene_txt = f.read()
    runtime_container = get_runtime_container(scene, scene_txt)
    # return runtime_container.get_object_list_with_children_as_string()
    ret = runtime_container.as_llmjson()
    del ret["robot0"]
    return ret


def get_runtime_container(scene, scene_txt_for_memoization):
    print("CACHE MISS: getting runtime container...")

    runtime_container = get_sim(scene, just_runtime_container=True)

    return runtime_container


def resolve_scene_id(floor_name):
    split, house, _ = path_2_parts(floor_name)
    import prior

    dataset = prior.load_dataset("procthor-10k")
    floor_name = dataset[split][house]

    if isinstance(floor_name, dict):
        return floor_name

    if isinstance(floor_name, int) or floor_name.startswith("FloorPlan"):
        floor_name = str(floor_name)
        return f"FloorPlan{floor_name.replace('FloorPlan', '')}"

    if isinstance(floor_name, str) and "procthor" in floor_name:
        return floor_name

    if not floor_name.endswith(".json"):
        if os.path.exists(floor_name + "/scene.json"):
            floor_name += "/scene.json"

    assert floor_name.endswith(".json")

    return floor_name

    with open(floor_name, "r") as f:
        scene = json.load(f)
    return scene


def remove_repeated_pickupables(controller, pickupables) -> Dict[str, Any]:
    objects_id = [obj["objectId"] for obj in controller.last_event.metadata["objects"]]
    pickupable_name = [p_name.split("|")[0] for p_name in pickupables]
    for obj_id in objects_id:
        # Check if the prefix is in the list of pickupable
        if obj_id.split("|")[0] in pickupable_name and obj_id not in pickupables:
            print(f"Disabling repeated pickupable: {obj_id}")
            controller.step(action="DisableObject", objectId=obj_id)
            # README: for some reason this line hangs...
            # controller.step(action="RemoveFromScene", objectId=obj_id)


def removed_pickupables(controller, target_pickupables) -> Dict[str, Any]:
    for obj in controller.last_event.metadata["objects"]:
        obj_id = obj["objectId"]
        is_pickupable = obj["pickupable"]
        if is_pickupable and obj_id not in target_pickupables:
            print(f"Disabling non-pickupable: {obj_id}")
            controller.step(action="DisableObject", objectId=obj_id)


def random_teleport(controller: Controller, full_reachability_graph: ReachabilityGraph):
    possible_start_nodes = list(full_reachability_graph.nodes)
    possible_start_nodes = list(sorted(possible_start_nodes))
    random_pos = random.choice(possible_start_nodes)
    print(f"Teleporting to random position: {random_pos}")
    controller.step(
        action="TeleportFull",
        position={
            "x": random_pos[0],
            "y": DEFAULT_ROBOT_HEIGHT,
            "z": random_pos[2],
        },
        rotation=DEFAULT_ROBOT_ROT,
        standing=True,
        horizon=30,
        forceAction=True,
    )


def get_sim(
    floor_no,
    just_controller=False,
    just_controller_no_setup=False,
    humanviewing_params={},
    renderInstanceSegmentation=False,
    renderDepthImage=False,
    im_width: int = 1250,
    im_height: int = 1250,
    fov: int = 90,
    grid_size: float = 0.25,
    save_topdown: bool = True,
    mindist_to_walls: float = 0.3,
    mindist_to_obj: float = 0.01,
):
    try:
        generated_sss_data = load_sssd(floor_no)
        # Seed scene
        _, _, seed = path_2_parts(floor_no)
        seed_everything(seed)
    except Exception as e:
        print("Error while loading SSD:", e)
        generated_sss_data = None
        seed = 0
        seed_everything(seed)

    if not isinstance(floor_no, dict):
        scene = resolve_scene_id(floor_no)
        assert sum(list(map(int, (just_controller, just_controller_no_setup)))) <= 1
    else:
        scene = floor_no

    controller = get_hippo_controller(
        scene,
        width=im_width,
        height=im_height,
        snapGrid=False,
        snapToGrid=False,
        visibilityDistance=1,
        fieldOfView=fov,
        gridSize=grid_size,
        rotateStepDegrees=20,
    )

    if just_controller_no_setup:
        return controller

    # Remove repeated pickupables so they are unique
    # remove_repeated_pickupables(controller, generated_sss_data.pickupable_names)
    if generated_sss_data is not None:
        removed_pickupables(controller, generated_sss_data.pickupable_names)
    all_initial_objects = controller.last_event.metadata["objects"]
    NO_ROBOTS = 1

    CEILING_HEIGHT = controller.last_event.metadata["sceneBounds"]["size"]["y"]

    from semistaticsim.rendering.floor import FloorPolygon

    floor_polygon = FloorPolygon.create(scene, grid_size)

    reachable_positions = controller.step("GetReachablePositions").metadata["actionReturn"]
    reachable_positions = [(p["x"], p["y"], p["z"]) for p in reachable_positions]
    # floor_polygon.plot_self(samples=reachable_positions)

    full_reachability_graph = floor_polygon.rechability_graph(
        pre_selected_points=reachable_positions,
        mindist_to_walls=mindist_to_walls,
        DEFAULT_ROBOT_HEIGHT=DEFAULT_ROBOT_HEIGHT,
        set_mindist_to_obj=mindist_to_obj,
    )

    # initialize n agents into the scene
    controller.step(
        dict(
            action="Initialize",
            agentMode="default",
            snapGrid=False,
            snapToGrid=False,
            gridSize=grid_size,
            rotateStepDegrees=90,
            visibilityDistance=1,
            fieldOfView=90,
            agentCount=NO_ROBOTS,
            renderInstanceSegmentation=renderInstanceSegmentation,
            renderDepthImage=renderDepthImage,
        ),
        raise_for_failure=True,
    )
    
    random_teleport(controller, full_reachability_graph)

    # run a few physics steps to make bad objects fall through floor...
    for _ in trange(10):
        controller.step(
            action="MoveAhead",
            moveMagnitude=0.01,
        )
    for obj in controller.last_event.metadata["objects"]:
        if obj["position"]["y"] < -1:
            controller.step(action="SetMassProperties", objectId=obj["objectId"], mass=0.0, drag=1500, angularDrag=1500)
            y = None
            for otherobj in all_initial_objects:
                if otherobj["objectId"] == obj["objectId"]:
                    y = otherobj["position"]["y"]
                    break
            assert y is not None
            controller.step(
                action="PlaceObjectAtPoint",
                objectId=obj["objectId"],
                position={
                    "x": obj["position"]["x"],
                    "y": y,
                    "z": obj["position"]["z"],
                },
            )
    random_teleport(controller, full_reachability_graph)

    # setting up tools for human viewing
    from semistaticsim.rendering.simulation.humanviewing import HumanViewing

    controller.humanviewing = HumanViewing(controller, None, **humanviewing_params)

    if just_controller:
        return controller

    simulator = Simulator(
        controller=controller,
        full_reachability_graph=full_reachability_graph,
        floor=floor_polygon,
        sss_data=generated_sss_data,
        save_topdown=save_topdown,
        seed=seed + 12345,
    )

    hu = HumanViewing(controller, simulator, **humanviewing_params)
    controller.humanviewing = hu
    simulator.humanviewing = hu

    print("Done building simulator")
    return simulator



import cv2
import gzip
import json
import math
import os
from typing import Sequence
from tqdm import tqdm
import numpy as np


actionList = {
    "MoveAhead": "w",
    "MoveBack": "s",
    "MoveLeft": "a",
    "MoveRight": "d",
    "RotateLeft": "q",
    "RotateRight": "e",
    "LookUp": "r",
    "LookDown": "f",
    "Record": "t",
    "FINISH": "p",
    # Interact action
    "PickupObject": "z",
    "PutObject": "x",
    "OpenObject": "c",
    "CloseObject": "v",
    "ToggleObjectOn": "b",
    "ToggleObjectOff": "n",
    "SliceObject": "m",
}


def get_interact_object(env):
    candidates = []
    objectId = ''

    for obj in env.last_event.metadata["objects"]:
        if obj["visible"]:
            candidates.append(obj["objectId"])

    if len(candidates) == 0:
        print('no valid interact object candidates')
        return None
    else:
        print('===========choose index from the candidates==========')

        # Create mapping from characters to indices
        char_to_index = {}
        index = 0

        # Add digits 0-9
        for i in range(10):
            if index < len(candidates):
                char_to_index[str(i)] = index
                print(f'{i} : {candidates[index]}')
                index += 1

        # Add lowercase letters a-z (will also handle uppercase via normalization)
        for c in 'abcdefghijklmnopqrstuvwxyz':
            if index < len(candidates):
                char_to_index[c] = index
                char_to_index[c.upper()] = index  # Map both cases to same index
                print(f'{c}/{c.upper()} : {candidates[index]}')
                index += 1

        # Add remaining uppercase letters if we have more than 36 objects
        for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            if index < len(candidates) and c not in char_to_index:
                char_to_index[c] = index
                print(f'{c} : {candidates[index]}')
                index += 1

        # Define shift key codes (common shift key codes)
        SHIFT_CODES = {225, 65505, 65506}  # Left Shift, Right Shift, etc.

        while True:
            keystroke = cv2.waitKey()
            print(f"Raw keystroke: {keystroke}")

            if keystroke == actionList["FINISH"]:
                print("stop interact")
                break

            # Ignore shift keys
            if keystroke in SHIFT_CODES:
                print("Ignoring Shift key...")
                continue

            # Convert keystroke to character
            try:
                # Convert ASCII code to character
                if 0 <= keystroke <= 255:
                    char = chr(keystroke)
                    print(f"Converted to char: '{char}'")

                    # Normalize to lowercase for letters (since we map both cases)
                    if char.isalpha():
                        normalized_char = char.lower()
                    else:
                        normalized_char = char

                    # Check if character is in our mapping
                    if normalized_char in char_to_index:
                        object_index = char_to_index[normalized_char]
                        objectId = candidates[object_index]
                        print(f"Selected object: {objectId}")
                        break
                    else:
                        print(f"INVALID KEY '{char}' - not in available options")
                else:
                    print(f"INVALID KEYCODE {keystroke}")
            except Exception as e:
                print(f"ERROR processing keystroke: {e}")
                continue

        return objectId

def keyboard_play(env, top_down_frames, first_view_frames, is_rotate, rotate_per_frame, HU_PLAN=[]):
    first_view_frame = env.last_event.frame
    #cv2.imshow("first_view", cv2.cvtColor(first_view_frame, cv2.COLOR_RGB2BGR))
    env.humanviewing.set_plan(HU_PLAN)
    env.humanviewing.incr_action_idx()
    env.humanviewing.display_augmented_frame()


    # remove the ceiling
    env.step(action="ToggleMapView")
    top_down_frame = env.last_event.third_party_camera_frames[0]
    cv2.imshow("top_view", cv2.cvtColor(top_down_frame, cv2.COLOR_RGB2BGR))
    env.step(action="ToggleMapView")

    step = 0

    while True:
        keystroke = cv2.waitKey(0)
        step += 1

        if keystroke == ord(actionList["FINISH"]):
            env.stop()
            cv2.destroyAllWindows()
            print("action: STOP")
            break

        if keystroke == ord(actionList["MoveAhead"]):
            action = "MoveAhead"
            print("action: MoveAhead")
        elif keystroke == ord(actionList["MoveBack"]):
            action = "MoveBack"
            print("action: MoveBack")
        elif keystroke == ord(actionList["MoveLeft"]):
            action = "MoveLeft"
            print("action: MoveLeft")
        elif keystroke == ord(actionList["MoveRight"]):
            action = "MoveRight"
            print("action: MoveRight")
        elif keystroke == ord(actionList["RotateLeft"]):
            action = "RotateLeft"
            print("action: RotateLeft")
        elif keystroke == ord(actionList["RotateRight"]):
            action = "RotateRight"
            print("action: RotateRight")
        elif keystroke == ord(actionList["LookUp"]):
            action = "LookUp"
            print("action: LookUp")
        elif keystroke == ord(actionList["LookDown"]):
            action = "LookDown"
            print("action: LookDown")

        elif keystroke == ord(actionList["PickupObject"]):
            action = "PickupObject"
            objectId = get_interact_object(env)
            pickup = objectId.split('|')[0]
            print('holding', pickup)
            print("action: PickupObject")
        elif keystroke == ord(actionList["PutObject"]):
            action = "PutObject"
            objectId = get_interact_object(env)
            print('putting on:', objectId)
            print("action: PutObject")
        elif keystroke == ord(actionList["OpenObject"]):
            action = "OpenObject"
            objectId = get_interact_object(env)
            print("action: OpenObject")
        elif keystroke == ord(actionList["CloseObject"]):
            action = "CloseObject"
            objectId = get_interact_object(env)
            print("action: CloseObject")
        elif keystroke == ord(actionList["ToggleObjectOn"]):
            action = "ToggleObjectOn"
            objectId = get_interact_object(env)
            print("action: ToggleObjectOn")
        elif keystroke == ord(actionList["ToggleObjectOff"]):
            action = "ToggleObjectOff"
            objectId = get_interact_object(env)
            print("action: ToggleObjectOff")
        elif keystroke == ord(actionList["SliceObject"]):
            action = "SliceObject"
            objectId = get_interact_object(env)
            print("action: SliceObject")
        else:
            print("INVALID KEY", keystroke)
            continue

        print("Interact object:")
        try:
            print(objectId)
        except UnboundLocalError:
            pass

        if action.startswith("Move"):
            pass
        else:
            env.humanviewing.incr_action_idx()

        # agent step
        if "Object" in action:
            env.step(action=action, objectId=objectId, forceAction=True)
        else:
            env.step(action=action)
        print(env.last_action)

        if is_rotate:
            ## rotation third camera
            pose = compute_rotate_camera_pose(env.last_event.metadata["sceneBounds"]["center"],
                                              env.last_event.metadata["thirdPartyCameras"][0], rotate_per_frame)

            del pose["agentPositionRelativeThirdPartyCameraPosition"]
            del pose["agentPositionRelativeThirdPartyCameraRotation"]

            env.step(
                action="UpdateThirdPartyCamera",
                **pose
            )

        first_view_frame = env.last_event.frame
        #cv2.imshow("first_view", cv2.cvtColor(first_view_frame, cv2.COLOR_RGB2BGR))
        env.humanviewing.display_augmented_frame()

        # remove the ceiling
        env.step(action="ToggleMapView")
        top_down_frame = env.last_event.third_party_camera_frames[0]
        cv2.imshow("top_view", cv2.cvtColor(top_down_frame, cv2.COLOR_RGB2BGR))
        env.step(action="ToggleMapView")

        top_down_frames.append(top_down_frame)
        first_view_frames.append(first_view_frame)


def show_video(frames: Sequence[np.ndarray], fps: int = 10):
    """Show a video composed of a sequence of frames.

    Example:
    frames = [
        controller.step("RotateRight", degrees=5).frame
        for _ in range(72)
    ]
    show_video(frames, fps=5)
    """
    from moviepy import ImageSequenceClip
    frames = ImageSequenceClip(frames, fps=fps)
    return frames


def export_video(path, frames):
    """Merges all the saved frames into a .mp4 video and saves it to `path`"""

    video = cv2.VideoWriter(
        path,
        cv2.VideoWriter_fourcc(*'mp4v'),
        5,
        (frames[0].shape[1], frames[0].shape[0]),
    )
    for frame in frames:
        # assumes that the frames are RGB images. CV2 uses BGR.
        video.write(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
    cv2.destroyAllWindows()
    video.release()


def compute_rotate_camera_pose(center, pose, degree_per_frame=6):
    """degree_per_frame: set the degree of rotation for each frame"""

    def rotate_pos(x1, y1, x2, y2, degree):
        angle = math.radians(degree)
        n_x1 = (x1 - x2) * math.cos(angle) - (y1 - y2) * math.sin(angle) + x2
        n_y1 = (x1 - x2) * math.sin(angle) + (y1 - y2) * math.cos(angle) + y2

        return n_x1, n_y1

    # print(math.sqrt((pose["position"]["x"]-center["x"])**2 + (pose["position"]["z"]-center["z"])**2))
    x, z = rotate_pos(pose["position"]["x"], pose["position"]["z"], center["x"], center["z"], degree_per_frame)
    pose["position"]["x"], pose["position"]["z"] = x, z

    direction_x = center["x"] - x
    direction_z = center["z"] - z
    pose["rotation"]["y"] = math.degrees(math.atan2(direction_x, direction_z))

    # print(math.sqrt((pose["position"]["x"]-center["x"])**2 + (pose["position"]["z"]-center["z"])**2))

    return pose


def initialize_side_camera_pose(scene_bound, pose, third_fov=60, slope_degree=45, down_angle=70, scale_factor=8):
    """
    down_angle: the x-axis rotation angle of the camera, represents the top view of the front view from top to bottom, which needs to be less than 90 degrees
    ensure the line vector between scene's center & camera 's angel equal down_angle
    scale_factor scale the camera's view, make it larger ensure camera can see the whole scene
    """
    fov_rad = np.radians(third_fov)
    pitch_rad = np.radians(down_angle)
    distance = (scene_bound["center"]["y"] / 2) / np.tan(fov_rad / 2)
    pose["position"]["y"] = scene_bound["center"]["y"] + distance * scale_factor * np.sin(pitch_rad)
    pose["position"]["z"] = scene_bound["center"]["z"] - distance * scale_factor * np.cos(pitch_rad)

    pose["rotation"]["x"] = down_angle

    pose["orthographic"] = False
    del pose["orthographicSize"]

    pose = compute_rotate_camera_pose(scene_bound["center"], pose, slope_degree)

    return pose


def main(scene_name="FloorPlan205_physics", gridSize=0.25, rotateStepDegrees=15,
         BEV=False, slope_degree=45, down_angle=65, use_procthor=False, procthor_scene_file="", procthor_scene_num=100,
         is_rotate=True, rotate_per_frame=6, generate_video=False, generate_gif=False, HU_PLAN=[]):
    ## procthor room
    if use_procthor:
        with gzip.open(procthor_scene_file, "r") as f:
            houses = [line for line in tqdm(f, total=10000, desc=f"Loading train")]
        ## procthor train set's room
        house = json.loads(houses[procthor_scene_num])
    else:
        ## select room, 1-30，201-230，301-330，401-430 are ithor's room
        house = scene_name

    #from SMARTLLM.smartllm.utils.resolve_scene import resolve_scene_id

    #floor_no = "/home/charlie/Desktop/Holodeck/SMARTLLM/hipposcenes/9/scene.json"
    #scene = resolve_scene_id(floor_no)
    #from SMARTLLM.smartllm.utils.get_controller import get_controller
    #controller = get_controller(scene, get_runtime_container=False, width=1000, height=1000, snapToGrid=False,
    #                                      visibilityDistance=100, fieldOfView=90, gridSize=0.25, rotateStepDegrees=20)

    from semistaticsim.rendering import get_sim
    controller = get_sim(scene_name, just_controller=True)
    """
    _ = Controller(
        agentMode="default",
        visibilityDistance=5,
        renderInstanceSegmentation=True,
        scene=house,
        # step sizes
        gridSize=gridSize,
        snapToGrid=False,
        rotateStepDegrees=rotateStepDegrees,
        # camera properties
        width=1200,
        height=800,
        fieldOfView=90,
        platform=CloudRendering,
    )
    """

    ## add third view camera
    event = controller.step(action="GetMapViewCameraProperties")
    ## third camera's fov
    third_fov = 60

    if not BEV:
        ## top_view(slope)
        pose = initialize_side_camera_pose(event.metadata["sceneBounds"], event.metadata["actionReturn"], third_fov,
                                           slope_degree, down_angle)
    else:
        ## BEV
        pose = event.metadata["actionReturn"]
        is_rotate = False  ## assume that BEV do not need rotation

    event = controller.step(
        action="AddThirdPartyCamera",
        skyboxColor="black",
        fieldOfView=third_fov,
        **pose
    )

    ## collect frame
    first_view_frames = []
    third_view_frames = []

    ## use keyboard control agent
    keyboard_play(controller, third_view_frames, first_view_frames, is_rotate, rotate_per_frame, HU_PLAN)

    ## use frames generate video
    if generate_video:

        if not os.path.exists("./video"):
            os.mkdir("./video")

        export_video("./video/first_view_{}.mp4".format(scene_name), first_view_frames)
        export_video("./video/third_view_{}.mp4".format(scene_name), third_view_frames)

    ## use frames generate gif
    if generate_gif:

        if not os.path.exists("./gif"):
            os.mkdir("./gif")

        clip = show_video(third_view_frames, fps=5)
        clip.write_gif("./gif/third_view_{}.gif".format(scene_name))
        clip2 = show_video(first_view_frames, fps=5)
        clip2.write_gif("./gif/first_view_{}.gif".format(scene_name))


if __name__ == "__main__":
    HU_PLAN = """
    1. GoToObject("Yellow cube")
    2. PickUpObject("Yellow cube")
    3. PutObjectDown("Yellow cube", "Table")
    """

    main(scene_name="/Users/charlie/Projects/Holodeck/hippo/sampled_scenes/realscenes/i_mug_kitchen_surely4/TRELLIS-yes-mask_True-aspect weighted/in_order_0",  # FloorPlan19_physics ## room
         gridSize=0.25, rotateStepDegrees=15,  ## agent step len and rotate degree
         BEV=False,  ## Bird's-eye view or top view(slope)
         slope_degree=60,  ## top view(slope)'s initial rotate degree
         down_angle=65,  ## top view(slope)'s pitch angle, should be 0-90, 90 equal to Bird's-eye view
         use_procthor=False,  ## use procthor room, True: select room from procthor train set, need dataset dir
         procthor_scene_file="",  ## procthor train set dir
         procthor_scene_num=100,  ## select scene from procthor train set
         is_rotate=True,  ## top_view rotate?
         rotate_per_frame=6,  ## top_view rotate degree
         generate_video=True,  ## use frames generate video
         generate_gif=True,  ## use frames generate gif
         HU_PLAN=HU_PLAN
         )


import dataclasses
from typing import Tuple, Union, List, Dict

import numpy as np
from ai2thor.controller import Controller

from semistaticsim.rendering.simulation.ai2thor_metadata_reader import get_robot_inventory, thunk_fix_robot_pos
from semistaticsim.rendering.simulation.skill_simulator import Simulator


@dataclasses.dataclass
class HumanViewing:
    c: Controller
    s: Union[Simulator, None]
    plan: list[str] = dataclasses.field(default_factory=lambda: [])
    current_action_idx: int = 0  # 0 = nothing yet, len(plan)+1 = all done
    use_altered_first_person_frame_instead: bool = False

    def set_plan(self, plan):
        if isinstance(plan, str):
            plan = list(filter(lambda x: len(x) > 0, map(str.strip, plan.split("\n"))))
        self.plan = plan
        return plan

    def incr_action_idx(self):
        self.current_action_idx = self.current_action_idx + 1

    def get_latest_robot_frame(self):
        first_view_frame = self.c.last_event.frame
        return first_view_frame

    def get_latest_segmented_robot_frame(self):
        segmented_frame = self.c.last_event.instance_segmentation_frame
        return segmented_frame

    def get_latest_depth_robot_frame(self):
        depth_frame = self.c.last_event.depth_frame
        # Add last dimension to avoid plotting issues
        return depth_frame.reshape((depth_frame.shape[0], depth_frame.shape[1], 1))

    def get_latest_visible_instances(self, valid_instances: List[str]) -> List[str]:
        # Filter out doors, paintings, windows, walls, rooms
        f_do_not_consider = lambda x: any(y in x.lower() for y in ["door", "painting", "window", "wall", "room", "floor"])
        instance_ids = [obj_id for obj_id in self.c.last_event.instance_masks.keys() if not f_do_not_consider(obj_id)]
        # Only keep those instances in valid_receptacles
        instance_ids = [obj_id for obj_id in instance_ids if obj_id in valid_instances]
        return instance_ids

    def get_mapping_color_to_id(self) -> Dict[Tuple[float, float, float], str]:
        color_to_id = self.c.last_event.color_to_object_id
        return color_to_id
    
    def get_mapping_id_to_color(self) -> Dict[str, Tuple[float, float, float]]:
        id_to_color = self.c.last_event.object_id_to_color
        return id_to_color

    def get_camera_intrinsics(self) -> Dict[str, float]:
        im_width = self.c.last_event.metadata["screenWidth"]
        im_height = self.c.last_event.metadata["screenHeight"]
        fov = self.c.last_event.metadata["fov"]
        focal_length = (im_width / 2) / np.tan(np.deg2rad(fov) / 2)
        # Camera intrinsics
        fx, fy, cx, cy = focal_length, focal_length, im_width / 2, im_height / 2
        return {
            "image_width": im_width,
            "image_height": im_height,
            "fx": fx,
            "fy": fy,
            "cx": cx,
            "cy": cy,
            "fov_degrees": fov,
        }

    def get_agent_pose(self) -> Tuple[float, float, float]:
        thunk_fix_robot_pos(self.c)
        pos = self.c.last_event.metadata["agent"]["position"]
        rot = self.c.last_event.metadata["agent"]["rotation"]
        rot['x'] = self.c.last_event.metadata["agent"]["cameraHorizon"]
        return {
            'position': pos,
            'rotation': rot,
        }

    def get_latest_topdown_frame(self):
        #frame = self.c.last_event.third_party_camera_frames[0]
        self.c.step(action="ToggleMapView")
        frame = self.c.last_event.frame
        self.c.step(action="ToggleMapView")
        return frame

    def get_latest_altered_robot_frame(self):
        return self.c.last_event.third_party_camera_frames[1]

    def get_augmented_robot_frame_OLD(self, frame, message=None, held_item=None, hud_scale=2):
        import cv2, re

        if held_item is None:
            inventory = get_robot_inventory(self.c, 0)
            assert len(inventory) <= 1
            if len(inventory) == 0:
                held_item = None
            else:
                raw_item = inventory[0]
                held_item = re.match(r"^([^-]+)", raw_item).group(1)

        if message is None:
            if self.s is not None:
                message_queue = self.s.message_queue
                if len(message_queue) <= 0:
                    message = ""
                    # if self.s.currently_thinking:
                    #    message = "Judge currently thinking..."
                else:
                    message = message_queue[-1]
        if message is None or message == "":
            message = "[None yet]"

        hud_frame = frame.copy()
        h, w = hud_frame.shape[:2]

        def draw_box(img, top_left, bottom_right, color=(0, 0, 0), alpha=0.5):
            overlay = img.copy()
            cv2.rectangle(overlay, top_left, bottom_right, color, -1)
            return cv2.addWeighted(overlay, alpha, img, 1 - alpha, 0)

        def wrap_text(text, font, font_scale, thickness, max_width):
            """Split text into lines so each fits max_width, breaking long words if needed"""
            words = text.split(" ")
            lines, current = [], ""

            def measure_width(txt):
                (tw, _), _ = cv2.getTextSize(txt, font, font_scale, thickness)
                return tw

            for word in words:
                test_line = word if current == "" else current + " " + word
                if measure_width(test_line) <= max_width:
                    current = test_line
                else:
                    if current:  # push current line
                        lines.append(current)
                    # Now handle the long word (break into chunks if needed)
                    if measure_width(word) <= max_width:
                        current = word
                    else:
                        chunk = ""
                        for ch in word:
                            if measure_width(chunk + ch) <= max_width:
                                chunk += ch
                            else:
                                lines.append(chunk)
                                chunk = ch
                        current = chunk  # remainder stays in current
            if current:
                lines.append(current)
            return lines

        pad = int(20 * hud_scale)
        font = cv2.FONT_HERSHEY_SIMPLEX
        font_scale = 0.5 * hud_scale
        title_font_scale = 0.5 * hud_scale
        thickness = max(1, int(1 * hud_scale))

        # Calculate section widths (1/3 of the viewer each)
        section_width = w // 3
        inner_pad = pad // 2  # Padding inside each section

        # ---------------- Plan Display (Left section) ----------------
        plan_x = inner_pad
        plan_y = inner_pad
        title_plan = "Plan:"
        title_plan_size = cv2.getTextSize(title_plan, font, title_font_scale, thickness)[0]

        plan_box_width = section_width - 2 * inner_pad

        # Wrap each step individually
        wrapped_plan_steps = []
        for idx, step in enumerate(self.plan or ["[No plan yet]"], start=1):
            prefix = f""
            (prefix_width, _), _ = cv2.getTextSize(prefix, font, font_scale, thickness)

            # Wrap with prefix width reserved on the first line
            lines = wrap_text(step, font, font_scale, thickness, plan_box_width - inner_pad * 2 - prefix_width)

            if lines:
                lines[0] = prefix + lines[0]
            else:
                lines = [prefix]

            wrapped_plan_steps.extend(lines)

        plan_box_height = title_plan_size[1] + (len(wrapped_plan_steps) * int(20 * hud_scale)) + 2 * inner_pad

        plan_box_tl = (plan_x, plan_y + title_plan_size[1] + inner_pad)
        plan_box_br = (plan_box_tl[0] + plan_box_width, plan_box_tl[1] + plan_box_height)
        hud_frame = draw_box(hud_frame, plan_box_tl, plan_box_br, (50, 50, 0), 0.7)

        cv2.putText(
            hud_frame,
            title_plan,
            (plan_x, plan_y + title_plan_size[1]),
            font,
            title_font_scale,
            (0, 0, 0),
            thickness,
            cv2.LINE_AA,
        )

        line_y = plan_box_tl[1] + inner_pad + title_plan_size[1]
        step_idx = None
        for line in wrapped_plan_steps:
            match = re.match(r"^(\d+)\.", line)
            if match:
                step_idx = int(match.group(1))

            color = (180, 180, 180)
            if step_idx is not None:
                if self.current_action_idx == step_idx:
                    color = (0, 255, 255)
                elif self.current_action_idx > step_idx:
                    color = (255, 140, 0)
            else:
                color = (255, 140, 0)

            cv2.putText(hud_frame, line, (plan_x + inner_pad, line_y), font, font_scale, color, thickness, cv2.LINE_AA)
            line_y += int(20 * hud_scale)

            # Advance step index only when this line is the first of a step
            # if line.startswith(f"{step_idx}. "):
            #    step_idx += 1

        # ---------------- System Message (Middle section) ----------------
        sys_x = section_width + inner_pad
        sys_y = inner_pad
        title_msg = "System message:"
        title_size = cv2.getTextSize(title_msg, font, title_font_scale, thickness)[0]

        msg_box_width = section_width - 2 * inner_pad
        message_lines = wrap_text(message, font, font_scale, thickness, msg_box_width - inner_pad * 2)
        msg_box_height = title_size[1] + (len(message_lines) * int(20 * hud_scale)) + 3 * inner_pad

        msg_box_tl = (sys_x, sys_y + title_size[1] + inner_pad)
        msg_box_br = (msg_box_tl[0] + msg_box_width, msg_box_tl[1] + msg_box_height)
        hud_frame = draw_box(hud_frame, msg_box_tl, msg_box_br, (0, 0, 50), 0.7)

        cv2.putText(
            hud_frame,
            title_msg,
            (sys_x, sys_y + title_size[1]),
            font,
            title_font_scale,
            (0, 0, 0),
            thickness,
            cv2.LINE_AA,
        )

        line_y = msg_box_tl[1] + inner_pad + title_size[1]
        for line in message_lines:
            cv2.putText(
                hud_frame,
                line,
                (msg_box_tl[0] + inner_pad, line_y),
                font,
                font_scale,
                (255, 255, 255),
                thickness,
                cv2.LINE_AA,
            )
            line_y += int(20 * hud_scale)

        # ---------------- Held Item (Right section) ----------------
        item_x = 2 * section_width + inner_pad
        item_y = inner_pad
        title_item = "Held item:"
        item_text = f"[{held_item}]" if held_item else "[Empty Gripper]"
        title_item_size = cv2.getTextSize(title_item, font, title_font_scale, thickness)[0]

        held_object_image = self.c.get_segmented_held_object()
        if held_object_image is not None:
            held_object_image = cv2.cvtColor(held_object_image, cv2.COLOR_RGBA2BGRA)

        # Calculate the box dimensions
        item_box_width = section_width - 2 * inner_pad

        # If we have an image, adjust the box height to accommodate it
        if held_object_image is not None:
            # Calculate maximum dimensions for the image (80% of box width)
            max_img_width = int(item_box_width * 0.8)
            max_img_height = max_img_width  # Keep it square

            # Get original dimensions
            img_h, img_w = held_object_image.shape[:2]

            # Calculate scaling factor
            scale = min(max_img_width / img_w, max_img_height / img_h)

            # Resize the image
            new_width = int(img_w * scale)
            new_height = int(img_h * scale)
            resized_image = cv2.resize(held_object_image, (new_width, new_height))

            # Calculate text height needed
            text_height = title_item_size[1] + (1 * int(20 * hud_scale)) + 2 * inner_pad

            # Total box height = text height + image height + padding
            item_box_height = text_height + new_height + inner_pad

            # Create the box
            item_box_tl = (item_x, item_y + title_item_size[1] + inner_pad)
            item_box_br = (item_box_tl[0] + item_box_width, item_box_tl[1] + item_box_height)
            hud_frame = draw_box(hud_frame, item_box_tl, item_box_br, (50, 0, 0), 0.7)

            # Draw title and text
            cv2.putText(
                hud_frame,
                title_item,
                (item_x, item_y + title_item_size[1]),
                font,
                title_font_scale,
                (0, 0, 0),
                thickness,
                cv2.LINE_AA,
            )

            line_y = item_box_tl[1] + inner_pad + title_item_size[1]
            cv2.putText(
                hud_frame,
                item_text,
                (item_box_tl[0] + inner_pad, line_y),
                font,
                font_scale,
                (255, 255, 255),
                thickness,
                cv2.LINE_AA,
            )

            # Calculate image position (centered horizontally)
            img_x = item_box_tl[0] + (item_box_width - new_width) // 2
            img_y = line_y + int(20 * hud_scale) + inner_pad

            # Overlay the image on the HUD frame
            tmp = np.ones((hud_frame.shape[0], hud_frame.shape[1], 4), dtype=hud_frame.dtype)
            tmp[:, :, :3] = hud_frame
            hud_frame = tmp
            hud_frame[img_y : img_y + new_height, img_x : img_x + new_width] = resized_image
            hud_frame = hud_frame[:, :, :3]

        else:
            # No image - just show text as before
            item_lines = wrap_text(item_text, font, font_scale, thickness, item_box_width - inner_pad * 2)
            item_box_height = title_item_size[1] + (len(item_lines) * int(20 * hud_scale)) + 3 * inner_pad

            item_box_tl = (item_x, item_y + title_item_size[1] + inner_pad)
            item_box_br = (item_box_tl[0] + item_box_width, item_box_tl[1] + item_box_height)
            hud_frame = draw_box(hud_frame, item_box_tl, item_box_br, (50, 0, 0), 0.7)

            cv2.putText(
                hud_frame,
                title_item,
                (item_x, item_y + title_item_size[1]),
                font,
                title_font_scale,
                (0, 0, 0),
                thickness,
                cv2.LINE_AA,
            )

            line_y = item_box_tl[1] + inner_pad + title_item_size[1]
            for line in item_lines:
                cv2.putText(
                    hud_frame,
                    line,
                    (item_box_tl[0] + inner_pad, line_y),
                    font,
                    font_scale,
                    (255, 255, 255),
                    thickness,
                    cv2.LINE_AA,
                )
                line_y += int(20 * hud_scale)

        return hud_frame

    def get_augmented_robot_frame(self, frame, message=None, held_item=None, hud_scale=2):
        import cv2, re

        if held_item is None:
            inventory = get_robot_inventory(self.c, 0)
            assert len(inventory) <= 1
            if len(inventory) == 0:
                held_item = None
            else:
                raw_item = inventory[0]
                held_item = re.match(r"^([^-]+)", raw_item).group(1)

        if message is None:
            if self.s is not None:
                message_queue = self.s.message_queue
                if len(message_queue) <= 0:
                    message = ""
                    # if self.s.currently_thinking:
                    #    message = "Judge currently thinking..."
                else:
                    message = message_queue[-1]
        if message is None or message == "":
            message = "[None yet]"

        hud_frame = frame.copy()
        h, w = hud_frame.shape[:2]

        def draw_box(img, top_left, bottom_right, color=(0, 0, 0), alpha=0.5):
            overlay = img.copy()
            cv2.rectangle(overlay, top_left, bottom_right, color, -1)
            return cv2.addWeighted(overlay, alpha, img, 1 - alpha, 0)

        def wrap_text(text, font, font_scale, thickness, max_width):
            """Split text into lines so each fits max_width, breaking long words if needed"""
            words = text.split(" ")
            lines, current = [], ""

            def measure_width(txt):
                (tw, _), _ = cv2.getTextSize(txt, font, font_scale, thickness)
                return tw

            for word in words:
                test_line = word if current == "" else current + " " + word
                if measure_width(test_line) <= max_width:
                    current = test_line
                else:
                    if current:  # push current line
                        lines.append(current)
                    # Now handle the long word (break into chunks if needed)
                    if measure_width(word) <= max_width:
                        current = word
                    else:
                        chunk = ""
                        for ch in word:
                            if measure_width(chunk + ch) <= max_width:
                                chunk += ch
                            else:
                                lines.append(chunk)
                                chunk = ch
                        current = chunk  # remainder stays in current
            if current:
                lines.append(current)
            return lines

        pad = int(20 * hud_scale)
        font = cv2.FONT_HERSHEY_SIMPLEX
        base_font_scale = 0.5 * hud_scale
        title_font_scale = 0.5 * hud_scale
        thickness = max(1, int(1 * hud_scale))

        # Calculate section widths (1/3 of the viewer each)
        section_width = w // 3
        inner_pad = pad // 2  # Padding inside each section

        # ---------------- Plan Display (Left section) ----------------
        plan_x = inner_pad
        plan_y = inner_pad
        title_plan = "Plan:"
        title_plan_size = cv2.getTextSize(title_plan, font, title_font_scale, thickness)[0]

        plan_box_width = section_width - 2 * inner_pad
        max_plan_height = h - 2.5 * inner_pad  # Maximum allowed height for the plan section

        # Function to calculate plan height with given font scale
        def calculate_plan_height(font_scale):
            wrapped_plan_steps = []
            for idx, step in enumerate(self.plan or ["[No plan yet]"], start=1):
                prefix = f""
                (prefix_width, _), _ = cv2.getTextSize(prefix, font, font_scale, thickness)

                # Wrap with prefix width reserved on the first line
                lines = wrap_text(step, font, font_scale, thickness, plan_box_width - inner_pad * 2 - prefix_width)

                if lines:
                    lines[0] = prefix + lines[0]
                else:
                    lines = [prefix]

                wrapped_plan_steps.extend(lines)

            line_height = int(20 * hud_scale * (font_scale / base_font_scale))
            return title_plan_size[1] + (len(wrapped_plan_steps) * line_height) + 2 * inner_pad

        # Adjust font scale if plan is too long
        font_scale = base_font_scale
        min_font_scale = 0.1 * hud_scale  # Minimum font scale to prevent text from becoming unreadable

        plan_height = calculate_plan_height(font_scale)
        while plan_height > max_plan_height and font_scale > min_font_scale:
            font_scale -= 0.025
            plan_height = calculate_plan_height(font_scale)

        # Now generate the wrapped plan with the adjusted font scale
        wrapped_plan_steps = []
        for idx, step in enumerate(self.plan or ["[No plan yet]"], start=1):
            prefix = f""
            (prefix_width, _), _ = cv2.getTextSize(prefix, font, font_scale, thickness)

            # Wrap with prefix width reserved on the first line
            lines = wrap_text(step, font, font_scale, thickness, plan_box_width - inner_pad * 2 - prefix_width)

            if lines:
                lines[0] = prefix + lines[0]
            else:
                lines = [prefix]

            wrapped_plan_steps.extend(lines)

        line_height = int(20 * hud_scale * (font_scale / base_font_scale))
        plan_box_height = title_plan_size[1] + (len(wrapped_plan_steps) * line_height) + 2 * inner_pad

        plan_box_tl = (plan_x, plan_y + title_plan_size[1] + inner_pad)
        plan_box_br = (plan_box_tl[0] + plan_box_width, plan_box_tl[1] + plan_box_height)
        hud_frame = draw_box(hud_frame, plan_box_tl, plan_box_br, (50, 50, 0), 0.7)

        cv2.putText(
            hud_frame,
            title_plan,
            (plan_x, plan_y + title_plan_size[1]),
            font,
            title_font_scale,
            (0, 0, 0),
            thickness,
            cv2.LINE_AA,
        )

        line_y = plan_box_tl[1] + inner_pad + title_plan_size[1]
        step_idx = None
        for line in wrapped_plan_steps:
            match = re.match(r"^(\d+)\.", line)
            if match:
                step_idx = int(match.group(1))

            color = (180, 180, 180)
            if step_idx is not None:
                if self.current_action_idx == step_idx:
                    color = (0, 255, 255)
                elif self.current_action_idx > step_idx:
                    color = (255, 140, 0)
            else:
                color = (255, 140, 0)

            cv2.putText(hud_frame, line, (plan_x + inner_pad, line_y), font, font_scale, color, thickness, cv2.LINE_AA)
            line_y += line_height

        # ---------------- System Message (Middle section) ----------------
        sys_x = section_width + inner_pad
        sys_y = inner_pad
        title_msg = "System message:"
        title_size = cv2.getTextSize(title_msg, font, title_font_scale, thickness)[0]

        msg_box_width = section_width - 2 * inner_pad
        message_lines = wrap_text(message, font, base_font_scale, thickness, msg_box_width - inner_pad * 2)
        msg_box_height = title_size[1] + (len(message_lines) * int(20 * hud_scale)) + 3 * inner_pad

        msg_box_tl = (sys_x, sys_y + title_size[1] + inner_pad)
        msg_box_br = (msg_box_tl[0] + msg_box_width, msg_box_tl[1] + msg_box_height)
        hud_frame = draw_box(hud_frame, msg_box_tl, msg_box_br, (0, 0, 50), 0.7)

        cv2.putText(
            hud_frame,
            title_msg,
            (sys_x, sys_y + title_size[1]),
            font,
            title_font_scale,
            (0, 0, 0),
            thickness,
            cv2.LINE_AA,
        )

        line_y = msg_box_tl[1] + inner_pad + title_size[1]
        for line in message_lines:
            cv2.putText(
                hud_frame,
                line,
                (msg_box_tl[0] + inner_pad, line_y),
                font,
                base_font_scale,
                (255, 255, 255),
                thickness,
                cv2.LINE_AA,
            )
            line_y += int(20 * hud_scale)

        # ---------------- Held Item (Right section) ----------------
        item_x = 2 * section_width + inner_pad
        item_y = inner_pad
        title_item = "Held item:"
        item_text = f"[{held_item}]" if held_item else "[Empty Gripper]"
        title_item_size = cv2.getTextSize(title_item, font, title_font_scale, thickness)[0]

        held_object_image = self.c.get_segmented_held_object()
        if held_object_image is not None:
            held_object_image = cv2.cvtColor(held_object_image, cv2.COLOR_RGBA2BGRA)

        # Calculate the box dimensions
        item_box_width = section_width - 2 * inner_pad

        # If we have an image, adjust the box height to accommodate it
        if held_object_image is not None:
            # Calculate maximum dimensions for the image (80% of box width)
            max_img_width = int(item_box_width * 0.8)
            max_img_height = max_img_width  # Keep it square

            # Get original dimensions
            img_h, img_w = held_object_image.shape[:2]

            # Calculate scaling factor
            scale = min(max_img_width / img_w, max_img_height / img_h)

            # Resize the image
            new_width = int(img_w * scale)
            new_height = int(img_h * scale)
            resized_image = cv2.resize(held_object_image, (new_width, new_height))

            # Calculate text height needed
            text_height = title_item_size[1] + (1 * int(20 * hud_scale)) + 2 * inner_pad

            # Total box height = text height + image height + padding
            item_box_height = text_height + new_height + inner_pad

            # Create the box
            item_box_tl = (item_x, item_y + title_item_size[1] + inner_pad)
            item_box_br = (item_box_tl[0] + item_box_width, item_box_tl[1] + item_box_height)
            hud_frame = draw_box(hud_frame, item_box_tl, item_box_br, (50, 0, 0), 0.7)

            # Draw title and text
            cv2.putText(
                hud_frame,
                title_item,
                (item_x, item_y + title_item_size[1]),
                font,
                title_font_scale,
                (0, 0, 0),
                thickness,
                cv2.LINE_AA,
            )

            line_y = item_box_tl[1] + inner_pad + title_item_size[1]
            cv2.putText(
                hud_frame,
                item_text,
                (item_box_tl[0] + inner_pad, line_y),
                font,
                base_font_scale,
                (255, 255, 255),
                thickness,
                cv2.LINE_AA,
            )

            # Calculate image position (centered horizontally)
            img_x = item_box_tl[0] + (item_box_width - new_width) // 2
            img_y = line_y + int(20 * hud_scale) + inner_pad

            # Overlay the image on the HUD frame
            tmp = np.ones((hud_frame.shape[0], hud_frame.shape[1], 4), dtype=hud_frame.dtype)
            tmp[:, :, :3] = hud_frame
            hud_frame = tmp
            hud_frame[img_y : img_y + new_height, img_x : img_x + new_width] = resized_image
            hud_frame = hud_frame[:, :, :3]

        else:
            # No image - just show text as before
            item_lines = wrap_text(item_text, font, base_font_scale, thickness, item_box_width - inner_pad * 2)
            item_box_height = title_item_size[1] + (len(item_lines) * int(20 * hud_scale)) + 3 * inner_pad

            item_box_tl = (item_x, item_y + title_item_size[1] + inner_pad)
            item_box_br = (item_box_tl[0] + item_box_width, item_box_tl[1] + item_box_height)
            hud_frame = draw_box(hud_frame, item_box_tl, item_box_br, (50, 0, 0), 0.7)

            cv2.putText(
                hud_frame,
                title_item,
                (item_x, item_y + title_item_size[1]),
                font,
                title_font_scale,
                (0, 0, 0),
                thickness,
                cv2.LINE_AA,
            )

            line_y = item_box_tl[1] + inner_pad + title_item_size[1]
            for line in item_lines:
                cv2.putText(
                    hud_frame,
                    line,
                    (item_box_tl[0] + inner_pad, line_y),
                    font,
                    base_font_scale,
                    (255, 255, 255),
                    thickness,
                    cv2.LINE_AA,
                )
                line_y += int(20 * hud_scale)

        return hud_frame

    # def display_frame(self, frame):
    #    import cv2
    #    if frame is None:
    #        if self.use_altered_first_person_frame_instead:
    #            frame = self.get_latest_altered_robot_frame()
    #        else:
    #            frame = self.get_latest_robot_frame()
    #    cv2.imshow("first_view", cv2.cvtColor(frame, cv2.COLOR_RGB2BGR))

    def display_augmented_frame(self):
        if self.use_altered_first_person_frame_instead:
            frame = self.get_latest_altered_robot_frame()
        else:
            frame = self.get_latest_robot_frame()

        return self.display_frame(self.get_augmented_robot_frame(frame))



import os
import random
import re
import json
import numpy as np


def objects_to_txt():
    curdir = "/".join(__file__.split("/")[:-1])

    with open(os.path.join(curdir, "pickupables_prior.json")) as f:
        pickupable_prior = json.load(f)
    with open(os.path.join(curdir, "receptacles_prior.json")) as f:
        receptacle_prior = json.load(f)

    def split_camel_preserve_acronyms(name):
        # Insert space between lowercase → uppercase
        # OR between acronym → normal word
        s = re.sub(r"(?<=[a-z])(?=[A-Z])", " ", name)
        s = re.sub(r"(?<=[A-Z])(?=[A-Z][a-z])", " ", s)
        return s.lower()

    objects = set(pickupable_prior.keys()) | set(receptacle_prior.keys())
    objects = {split_camel_preserve_acronyms(obj) for obj in objects}

    # Save to txt file
    with open(os.path.join(curdir, "objects.txt"), "w") as f:
        for obj in sorted(objects):
            f.write(f"{obj}\n")


def seed_everything(seed: int = 42):
    random.seed(seed)
    np.random.seed(seed)
    # Make hashing deterministic
    os.environ["PYTHONHASHSEED"] = str(seed)

def is_sublevel(name: str) -> str:
    # Check if receptacle is a sublevel one by looking for '___' pattern
    return bool(re.search(r"___", name))

def format_receptacle(name: str) -> str:
    # Extract all digit sequences and join them
    numbers = "".join(re.findall(r"\d+", name))

    # Remove everything after last non-digit separator, then strip trailing '|' or whitespace
    label = re.split(r"\d+", name)[0].rstrip("| ").strip()

    return f"{label} # {numbers}" if numbers else label


if __name__ == "__main__":
    objects_to_txt()



import itertools
import json
import os
from typing import List, Dict, Optional, Any
import math

import numpy as np
from semistaticsim.groundtruth.utils import seed_everything
from tqdm import tqdm

from flax import struct
import jax.numpy as jnp

from semistaticsim.groundtruth.constants import AFFORDANCE_NAMES
from semistaticsim.groundtruth.procthor_object import ProcThorObject
from semistaticsim.groundtruth.utils import is_sublevel
# os.environ["JAX_PLATFORM_NAME"] = "cpu"
import jax

# jax.config.update('jax_platform_name', "cpu")
import prior

from semistaticsim.rendering.ai2thor_hippo_controller import get_sim
from semistaticsim.rendering.floor import ReachabilityGraph
from semistaticsim.rendering.utils.dict2xyztup import dict2xyztuple


def str2int(str):
    # Convert string to bytes
    b = str.encode("utf-8")  # or "ascii" if you know it's ASCII

    # Convert bytes to integer
    i = int.from_bytes(b, byteorder="big")  # use "little" for little-endian
    return i


def int2str(i):
    b_back = i.to_bytes((i.bit_length() + 7) // 8, byteorder="big")
    s_back = b_back.decode("utf-8")
    return s_back


def add_receptacle(receptacles: Dict[str, Dict[str, float]], new_receptacle, new_p):
    # Scale down all existing probabilities to make space for new_p
    for _, v in receptacles.items():
        v["p"] *= 1 - new_p

    # Add new object
    receptacles[new_receptacle] = {"p": new_p}

    # Optionally recompute counts if they should stay proportional
    total_count = sum(v["count"] for v in receptacles.values() if "count" in v)
    new_count = int(round(total_count * new_p))  # scale by total count
    receptacles[new_receptacle]["count"] = new_count

    # Renormalize counts to keep integer consistency
    total_count_new = total_count + new_count
    for v in receptacles.values():
        v["p"] = v["count"] / total_count_new

    return receptacles


def specific_id_to_generic(id: str) -> str:
    return id.split("|")[0]


@struct.dataclass
class ObjectCollection:
    pickupables: ProcThorObject  # vmapped
    receptacles: ProcThorObject  # vmapped
    mat_pickupable_to_receptacles: jnp.ndarray

    pickupable_ids: List[str] = struct.field(pytree_node=False)
    receptacle_ids: List[str] = struct.field(pytree_node=False)
    pickupable_to_receptacles: Dict[str, List[str]] = struct.field(pytree_node=False)

    @staticmethod
    def _sample_unique_pickupables(key, valid_pickupables, limit_num_pickupables, p_pickupables):
        """
        Sample pickupables ensuring no two pickupables have the exact same name.
        """
        seen_names = set()
        pickupables_to_keep = []

        # Create a list of available indices
        available_indices = list(range(len(valid_pickupables)))
        p_available = p_pickupables.copy()

        while len(pickupables_to_keep) < limit_num_pickupables and len(available_indices) > 0:
            # Renormalize probabilities
            p_normalized = p_available / p_available.sum()

            # Sample one index
            key, rng = jax.random.split(key)
            sampled_idx = jax.random.choice(rng, len(available_indices), replace=False, p=p_normalized).item()

            actual_idx = available_indices[sampled_idx]
            pickupable_name = specific_id_to_generic(valid_pickupables[actual_idx]["id"])

            # Check if we've already seen this name
            if pickupable_name not in seen_names:
                seen_names.add(pickupable_name)
                pickupables_to_keep.append(actual_idx)

            # Remove this index from available options
            available_indices.pop(sampled_idx)
            p_available = jnp.delete(p_available, sampled_idx)

        return pickupables_to_keep

    @classmethod
    def create(
        cls,
        path: str,
        key: jax.random.PRNGKey,
        do_surface_receptacles_only: bool = True,
        limit_num_pickupables: int = 2,
        limit_num_receptacles_per_pickupable: int = 3,
        max_pickupable_per_receptacle: int = 3,
        bias_sampling: bool = True,
        oob_receptacle_prob: float = 0.0,
    ):
        if not do_surface_receptacles_only:
            raise NotImplementedError()

        with open(f"{path}/receptacles.json", "r") as f:
            receptacles = json.load(f)

        with open(f"{path}/all_objects.json", "r") as f:
            all_objects = json.load(f)

        with open(f"{path}/pickupables.json", "r") as f:
            pickupables = json.load(f)

        with open(f"{path}/movables.json", "r") as f:
            moveables = json.load(f)

        curdir = "/".join(__file__.split("/")[:-1])
        with open(f"{curdir}/receptacles_prior.json") as f:
            receptacle_prior = json.load(f)
        with open(f"{curdir}/pickupables_prior.json") as f:
            pickupable_prior = json.load(f)

        # only consider non-openable receptacles
        INVALID_RECEPTACLES = jnp.nonzero(jnp.array([r["affordances"]["openable"] for r in receptacles]))[0].tolist()
        VALID_RECEPTACLES = jnp.nonzero(jnp.array([not r["affordances"]["openable"] for r in receptacles]))[0].tolist()

        # 1. Remove Pickupables with no available receptacles
        INVALID_PICKUPABLES = []
        VALID_PICKUPABLES = []
        for p in pickupables:
            for i, r_index in enumerate(p["valid_receptacles"]):
                if r_index not in VALID_RECEPTACLES:  # marks the bad receptacles as invalid
                    p["valid_receptacles"][i] = -1
            if (jnp.array(p["valid_receptacles"]) == -1).sum() < 2:
                INVALID_PICKUPABLES.append(p)
            else:
                VALID_PICKUPABLES.append(p)
        del pickupables

        # 2. Select some pickuables, mark the others as static
        if bias_sampling:
            p_pickupables = jnp.array(
                [pickupable_prior[specific_id_to_generic(p["id"])]["count"] for p in VALID_PICKUPABLES]
            )
            p_pickupables = p_pickupables / p_pickupables.sum()
        else:  # uniform sampling
            p_pickupables = jnp.ones((len(VALID_PICKUPABLES),)) / len(VALID_PICKUPABLES)
        key, rng = jax.random.split(key)
        limit_num_pickupables = len(VALID_PICKUPABLES) if limit_num_pickupables == -1 else limit_num_pickupables

        # Sample pickupables ensuring no duplicate names
        pickupables_to_keep = cls._sample_unique_pickupables(
            rng, VALID_PICKUPABLES, limit_num_pickupables, p_pickupables
        )
        for p_id in range(len(VALID_PICKUPABLES)):
            if p_id not in pickupables_to_keep:
                INVALID_PICKUPABLES.append(VALID_PICKUPABLES[p_id])
                VALID_PICKUPABLES[p_id] = None
        VALID_PICKUPABLES = list(filter(lambda x: x is not None, VALID_PICKUPABLES))

        # 3. Select some receptacles for each pickupable
        RECEPTACLES_FOR_THESE_PICKUPABLES = dict()
        receptacle_pickupable_count = {}  # Maps receptacle_id -> count of assigned pickupables

        # Scarcity-first ordering: process pickupables that have the FEWEST available receptacles first
        pickupable_order = sorted(
            range(len(VALID_PICKUPABLES)),
            key=lambda pid: sum(1 for r in VALID_PICKUPABLES[pid]["valid_receptacles"] if r != -1)
        )
        for p_id in pickupable_order:
            p = VALID_PICKUPABLES[p_id]

            receptacle_ids_for_this_p = [r for r in p["valid_receptacles"] if r != -1]

            # Filter out receptacles that have reached max_pickupable_per_receptacle
            available_receptacles = [
                r
                for r in receptacle_ids_for_this_p
                if max_pickupable_per_receptacle == -1 or receptacle_pickupable_count.get(r, 0) < max_pickupable_per_receptacle
            ]

            # If no receptacles are available (all are full), trigger error and cry
            if len(available_receptacles) == 0:
                raise RuntimeError(
                    f"All receptacles for pickupable {p['id']} have reached max capacity. Consider changing the seed or the scene."
                )

            if bias_sampling:
                r_names = [specific_id_to_generic(receptacles[i]["id"]) for i in available_receptacles]
                p_receptacles = jnp.array([receptacle_prior[specific_id_to_generic(name)]["count"] for name in r_names])
                p_receptacles = p_receptacles / p_receptacles.sum()
            else:
                p_receptacles = jnp.ones((len(available_receptacles),)) / len(available_receptacles)

            key, rng = jax.random.split(key)
            limit_num_receptacles_per_pickupable = (
                len(available_receptacles)
                if limit_num_receptacles_per_pickupable == -1
                else limit_num_receptacles_per_pickupable
            )
            max_receptacles = jax.random.randint(rng, (), 1, limit_num_receptacles_per_pickupable + 1)
            key, rng = jax.random.split(key)
            valid_receptacles_to_keep = jnp.sort(
                jax.random.choice(
                    rng,
                    jnp.array(available_receptacles),
                    (min(max_receptacles, len(available_receptacles)),),
                    replace=False,
                    p=p_receptacles,
                )
            ).tolist()
            RECEPTACLES_FOR_THESE_PICKUPABLES[p_id] = valid_receptacles_to_keep

            # Update the count for each assigned receptacle
            for r_id in valid_receptacles_to_keep:
                receptacle_pickupable_count[r_id] = receptacle_pickupable_count.get(r_id, 0) + 1
        ALL_RECEPTACLES_FOR_THESE_PICKUPABLES = list(
            set(itertools.chain.from_iterable(RECEPTACLES_FOR_THESE_PICKUPABLES.values()))
        )
        for p_id, p in enumerate(VALID_PICKUPABLES):
            for i in range(len(p["valid_receptacles"])):
                r_index = p["valid_receptacles"][i]
                if r_index not in RECEPTACLES_FOR_THESE_PICKUPABLES[p_id]:
                    p["valid_receptacles"][i] = -1
        for i in range(len(VALID_RECEPTACLES)):
            r = VALID_RECEPTACLES[i]
            if r not in ALL_RECEPTACLES_FOR_THESE_PICKUPABLES:
                VALID_RECEPTACLES[i] = None
                INVALID_RECEPTACLES.append(r)
        VALID_RECEPTACLES = list(filter(lambda x: x is not None, VALID_RECEPTACLES))

        # 4. Go and get the actual receptacles from their IDs
        VALID_RECEPTACLES = [receptacles[i] for i in VALID_RECEPTACLES]
        INVALID_RECEPTACLES = [receptacles[i] for i in INVALID_RECEPTACLES]

        # 4.5 todo rematerialize the indices to make all this faster (we only actually need VALID_RECEPTACLES, not all of them, but currently the indices in pickupables do consider all receptacles)
        if oob_receptacle_prob > 0:
            # Flip a coin for each pickupable to add the oob receptacle
            for p_id, data in RECEPTACLES_FOR_THESE_PICKUPABLES.items():
                key, rng = jax.random.split(key)
                coin = jax.random.uniform(rng, ())
                # Add OOB with 0.5 prob only if we have more than 1 receptacle already
                if len(data) > 1 and coin >= 0.5:
                    data.append(len(receptacles))
                    VALID_PICKUPABLES[p_id]["valid_receptacles"].append(len(receptacles))
                    print(f"Added OOB receptacle to pickupable {VALID_PICKUPABLES[p_id]['id']}")
                else:
                    VALID_PICKUPABLES[p_id]["valid_receptacles"].append(-1)
            SPAWN_COORD_FOR_OOB = VALID_RECEPTACLES[-1]["spawn_coordinates_above_receptacle"]
            SPAWN_COORD_FOR_OOB = [[x, -10, z] for (x, _, z) in SPAWN_COORD_FOR_OOB]
            OOB_FAKE_RECEPTACLE = {
                "id": "OOB_FAKE_RECEPTACLE",
                "position": {"x": 0, "y": -20, "z": 0},
                "rotation": {"x": 0, "y": 0, "z": 0},
                "aabb": VALID_RECEPTACLES[-1]["aabb"],
                "oobb": VALID_RECEPTACLES[-1]["oobb"],
                "affordances": VALID_RECEPTACLES[-1]["affordances"],
                "spawn_coordinates_above_receptacle": SPAWN_COORD_FOR_OOB,
                "valid_receptacles": [],
            }
            VALID_RECEPTACLES += [OOB_FAKE_RECEPTACLE]
            ALL_RECEPTACLES_FOR_THESE_PICKUPABLES.append(len(receptacles))
            receptacles += [OOB_FAKE_RECEPTACLE]

        # 4.5 rematerialize receptacles to remove inactive indices
        MAP_FULLRECEPTACLE_TO_VALIDRECEPTACLE_IDS = {j: i for i, j in enumerate(ALL_RECEPTACLES_FOR_THESE_PICKUPABLES)}
        VALID_RECEPTACLES = [receptacles[i] for i in ALL_RECEPTACLES_FOR_THESE_PICKUPABLES]

        # 5. save everything
        with open(f"{curdir}/pickupables.json") as f:
            pickupable_to_receptacle_weights = json.load(f)

        if oob_receptacle_prob > 0:
            for k in pickupable_to_receptacle_weights.keys():
                # Add obb receptacle prob to all pickupables so that the prob and counts match
                pickupable_to_receptacle_weights[k] = add_receptacle(
                    pickupable_to_receptacle_weights[k], "OOB_FAKE_RECEPTACLE", oob_receptacle_prob
                )

        mat_p_to_r = jnp.ones((len(VALID_PICKUPABLES), len(VALID_RECEPTACLES))) * -1
        p_name_to_r_names = {}
        # For each pickupable, get the valid receptacles and their weights
        for i, p in enumerate(VALID_PICKUPABLES):
            id = p["id"]
            valid_r_indices = p["valid_receptacles"]

            valid_r_names = [receptacles[i]["id"] for i in valid_r_indices if i != -1]
            valid_r_weights = np.array(
                [
                    pickupable_to_receptacle_weights[specific_id_to_generic(id)][specific_id_to_generic(r_id)]["p"]
                    for r_id in valid_r_names
                ]
            )
            valid_r_weights = (valid_r_weights / valid_r_weights.sum()).tolist()

            p_name_to_r_names[id] = valid_r_names
            for valid_r in valid_r_indices:
                if valid_r == -1:
                    continue
                mat_p_to_r = mat_p_to_r.at[i, MAP_FULLRECEPTACLE_TO_VALIDRECEPTACLE_IDS[valid_r]].set(
                    valid_r_weights.pop(0)
                )

        pickupable_names = [p["id"] for p in VALID_PICKUPABLES]
        receptacle_names = [r["id"] for r in VALID_RECEPTACLES]

        receptacle_instances = ProcThorObject.create(VALID_RECEPTACLES)
        pickupable_instances = ProcThorObject.create(VALID_PICKUPABLES)

        self = cls(
            pickupables=pickupable_instances,
            pickupable_ids=pickupable_names,
            receptacles=receptacle_instances,
            receptacle_ids=receptacle_names,
            pickupable_to_receptacles=p_name_to_r_names,
            mat_pickupable_to_receptacles=mat_p_to_r,
        )  # , moveable_instances)

        print("Object collection contains following mappings:")
        print(self.pickupable_to_receptacles)
        return self

    @staticmethod
    def _is_this_p_on_this_r(p, r):
        spawn_poses = r["spawn_coordinates_above_receptacle"]
        spawn_bbox = jnp.array(spawn_poses)
        min_xz, max_xz = jnp.min(spawn_bbox, axis=0)[0::2], jnp.max(spawn_bbox, axis=0)[0::2]

        p_pos_xz = jnp.array(dict2xyztuple(p["position"]))[0::2]  # gets first and last of the 3 (xyz)

        xz_is_ok = jnp.all((p_pos_xz >= min_xz) & (p_pos_xz <= max_xz))

        y_is_ok = jnp.array(dict2xyztuple(p["position"]))[1] > jnp.max(spawn_bbox, axis=0)[1]
        return jnp.logical_and(xz_is_ok, y_is_ok)
        # todo test if p_pos is on r
        # p_xy = jnp.array([p_pos[0], p_pos[2]])

    def valid_receptacle_indices_for_pickupable_index(self, p_i: int) -> jnp.ndarray:
        r_ids = self.mat_pickupable_to_receptacles[p_i]
        r_ids = r_ids > -1
        r_ids = jnp.nonzero(r_ids)[0]
        return r_ids

    def get_receptacle_by_index(self, index: int) -> ProcThorObject:
        return self.receptacles.take(index)

    def get_pickupable_by_index(self, index: int) -> ProcThorObject:
        return self.pickupables.take(index)

    def pickupable_id_to_index(self, id: str) -> int | None:
        for i, p in enumerate(self.pickupable_ids):
            if p == id:
                return i
        return None

    def receptacle_id_to_index(self, id: str) -> int | None:
        for i, r in enumerate(self.receptacle_ids):
            if r == id:
                return i
        return None

    def receptacle_index_to_id(self, i: int) -> str:
        return self.receptacle_ids[i]

    def pickupable_index_to_id(self, i: int) -> str:
        return self.pickupable_ids[i]

    def get_pickupable_by_id(self, i: int) -> str:
        return self.pickupable_ids[i]


def distance_to_box(point_pos, box_center, box_size):
    """
    Calculates the shortest distance from a point to an Axis-Aligned Bounding Box (AABB).



    Args:
        point_pos (tuple): (x, y, z) position of the point.
        box_center (tuple): (x, y, z) center of the box.
        box_size (tuple): (x, y, z) dimensions of the box (width, height, depth).

    Returns:
        float: The shortest Euclidean distance.
    """
    # Calculate the distance to the box on each axis.
    # The half_size is the distance from the center to the face of the box.
    half_size_x = box_size[0] / 2
    half_size_y = box_size[1] / 2
    half_size_z = box_size[2] / 2

    # The absolute distance from the point to the box center on each axis
    abs_dist_x = abs(point_pos[0] - box_center[0])
    abs_dist_y = abs(point_pos[1] - box_center[1])
    abs_dist_z = abs(point_pos[2] - box_center[2])

    # dx, dy, dz are the distances *outside* the box's extent on each axis.
    # If the point is inside or on the boundary on an axis, this value is 0.
    dx = max(0, abs_dist_x - half_size_x)
    dy = max(0, abs_dist_y - half_size_y)
    dz = max(0, abs_dist_z - half_size_z)

    # The total shortest distance is the Euclidean distance of (dx, dy, dz).
    return math.sqrt(dx * dx + dy * dy + dz * dz)


def write_object_affordances_for_scene(
    path: str, scene: Dict, receptacle_mappings: Dict, pickupable_mappings: Dict, pickupables_blacklist: list = None, unique_receptacles: bool = False
) -> None:
    sim = get_sim(
        scene,
        renderInstanceSegmentation=True,
        renderDepthImage=True,
        im_width=10,
        im_height=10,
        fov=90,
        save_topdown=False,
        mindist_to_walls=0.0,
        mindist_to_obj=0.0,
    )
    c = sim.controller  # Controller(scene=scene)
    reachability_graph: ReachabilityGraph = sim.full_reachability_graph

    objects = c.last_event.metadata["objects"]
    receptacles = []
    pickupables = []
    movables = []
    uninterestings = []

    def is_object_reachable(objId_or_obj):
        if isinstance(objId_or_obj, str):
            obj = [obj for obj in objects if obj["objectId"] == objId_or_obj][0]
        else:
            obj = objId_or_obj

        object_center = dict2xyztuple(obj["position"])
        object_size = dict2xyztuple(obj["axisAlignedBoundingBox"]["size"])
        node_position = reachability_graph.shunt(object_center)

        GRID_SIZE = reachability_graph.grid_size * 2
        ret = distance_to_box(node_position, object_center, object_size) < GRID_SIZE * 2
        return ret


    MAX_SIZE_SPAWN_COORDINATES = -np.inf

    for i, object in enumerate(objects):

        SKIP_THIS_OBJ = False
        for exclude in ["room", "door", "wall", "floor"]:
            if exclude in object["name"]:
                SKIP_THIS_OBJ = True
                break
        if SKIP_THIS_OBJ:
            continue

        id = object["objectId"]
        

        affordances = {affordance: object[affordance] for affordance in AFFORDANCE_NAMES}
        position = object["position"]
        rotation = object["rotation"]
        aabb = object["axisAlignedBoundingBox"]
        oobb = object["objectOrientedBoundingBox"]
        if oobb is None:
            oobb = {"cornerPoints": object["axisAlignedBoundingBox"]["cornerPoints"]}

        if id.split("|")[0] in receptacle_mappings and affordances["receptacle"]:
            spawnCoordinatesAboveReceptacle = c.step(
                action="GetSpawnCoordinatesAboveReceptacle", objectId=id, anywhere=True
            ).metadata["actionReturn"]
            spawnCoordinatesAboveReceptacle = [dict2xyztuple(p) for p in spawnCoordinatesAboveReceptacle]
        else:
            spawnCoordinatesAboveReceptacle = []

        MAX_SIZE_SPAWN_COORDINATES = max(MAX_SIZE_SPAWN_COORDINATES, len(spawnCoordinatesAboveReceptacle))

        ret = {
            "id": id,
            "position": position,
            "rotation": rotation,
            "aabb": aabb,
            "affordances": affordances,
            "oobb": oobb,
            "spawn_coordinates_above_receptacle": spawnCoordinatesAboveReceptacle,
        }

        obj_type = id.split("|")[0]

        if obj_type in receptacle_mappings and affordances["receptacle"] and is_object_reachable(id):
            if unique_receptacles:
                if not is_sublevel(id):
                    receptacles.append(ret)
                else:
                    uninterestings.append(ret)
            else:
                receptacles.append(ret)
        elif obj_type in pickupable_mappings and not affordances["receptacle"] and affordances["pickupable"] and obj_type not in pickupables_blacklist:
            pickupables.append(ret)
        elif affordances["moveable"]:
            movables.append(ret)
        else:
            uninterestings.append(ret)
    c.stop()

    for r in receptacles:
        assert len(r["spawn_coordinates_above_receptacle"]) > 0
        MISSING = MAX_SIZE_SPAWN_COORDINATES - len(r["spawn_coordinates_above_receptacle"])
        if MISSING > 0:
            arg_to_add = np.random.choice(len(r["spawn_coordinates_above_receptacle"]), MISSING, replace=True)
            to_add = np.array(r["spawn_coordinates_above_receptacle"])[arg_to_add]
            r["spawn_coordinates_above_receptacle"] = r["spawn_coordinates_above_receptacle"] + to_add.tolist()

    def find_receptacles_for_pickupable(clean_p_id: str, receptacle_list: List, mapping: Dict) -> List[int]:
        ret = []
        for receptacle_index, receptacle in enumerate(receptacle_list):
            clean_receptacle_id = receptacle["id"].split("|")[0]
            if clean_receptacle_id in mapping[clean_p_id]:
                ret.append(receptacle_index)
        MISSING = len(receptacles) - len(ret)
        ret += [-1] * MISSING
        return ret

    for p in pickupables:
        p_id = p["id"].split("|")[0]
        # Find the valid receptacles for this pickupable
        p["valid_receptacles"] = find_receptacles_for_pickupable(p_id, receptacles, pickupable_mappings)

    # for r in receptacles:
    # r_id = r["id"].split("|")[0]
    # Find the valid pickupables for this receptacle
    # r["valid_pickupables"] = find_receptacles_for_pickupable(r_id, pickupables, receptacle_mappings)

    def set_empty_receptacles(o: List[Dict[str, Any]]) -> None:
        if isinstance(o, list):
            return [set_empty_receptacles(o) for o in o]
        o["valid_receptacles"] = []

    def set_empty_pickupables(o: List[Dict[str, Any]]) -> None:
        if isinstance(o, list):
            return [set_empty_pickupables(o) for o in o]
        o["valid_pickupables"] = []

    set_empty_receptacles(receptacles)
    set_empty_pickupables(pickupables)
    set_empty_receptacles(uninterestings)
    set_empty_pickupables(uninterestings)
    set_empty_pickupables(movables)
    set_empty_receptacles(movables)

    os.makedirs(path, exist_ok=True)

    def save_objects(list, name):
        with open(path + f"/{name}.json", "w") as f:
            json.dump(list, f)

    save_objects(pickupables, "pickupables")
    save_objects(receptacles, "receptacles")
    save_objects(uninterestings, "uninterestings")
    save_objects(movables, "movables")
    save_objects(pickupables + receptacles + uninterestings + movables, "all_objects")


def write_object_affordances_for_procthor(path: str, unique_receptacles: bool = False, num_scenes_to_do: Optional[int] = None) -> None:
    dataset = prior.load_dataset("procthor-10k")
    dataset = {
        "test": dataset.test,
        "train": dataset.train,
        "val": dataset.val,
    }

    with open("/".join(__file__.split("/")[:-1]) + "/receptacles.json", "r") as f:
        receptacle_mappings = json.load(f)

    def pickupable_to_receptacle_prob(receptacles_dict: Dict) -> None:
        """
        Compute P(receptacle | pickupables) from counts in receptacles.json
        """
        pickupable_counts = {}  # object: total_count
        pickupable_to_receptacle = {}

        for receptacle, contained in receptacles_dict.items():
            for obj, stats in contained.items():
                count = stats["count"]
                pickupable_counts[obj] = pickupable_counts.get(obj, 0) + count

                if obj not in pickupable_to_receptacle:
                    pickupable_to_receptacle[obj] = {}
                pickupable_to_receptacle[obj][receptacle] = {"count": count}

        # Compute probabilities
        for obj, receptacle_map in pickupable_to_receptacle.items():
            total = pickupable_counts[obj]
            for receptacle, stats in receptacle_map.items():
                stats["p"] = stats["count"] / total if total > 0 else 0.0

        with open("/".join(__file__.split("/")[:-1]) + "/pickupables.json", "w") as f:
            json.dump(pickupable_to_receptacle, f, indent=4)

    def compute_prior_prob(receptacles_dict: Dict, f_name: str) -> None:
        """
        Compute P(receptacle) or P(pickupable) from counts in receptacles.json or pickupables.json
        """
        prior = {}
        total_count = 0

        for receptacle, contained in receptacles_dict.items():
            item_total = sum(stats["count"] for stats in contained.values())
            prior[receptacle] = {"count": item_total}
            total_count += item_total

        for _, stats in prior.items():
            stats["p"] = stats["count"] / total_count if total_count > 0 else 0.0

        with open(os.path.join("/".join(__file__.split("/")[:-1]), f_name), "w") as f:
            json.dump(prior, f, indent=4)

    # pickupable_to_receptacle_prob(receptacle_mappings)
    with open("/".join(__file__.split("/")[:-1]) + "/pickupables.json", "r") as f:
        pickupable_mappings = json.load(f)

    # Read pickupables blacklist
    with open("/".join(__file__.split("/")[:-1]) + "/pickupables_blacklist.json", "r") as f:
        blacklist = json.load(f)
    # compute_prior_prob(receptacle_mappings, "receptacles_prior.json")
    # compute_prior_prob(pickupable_mappings, "pickupables_prior.json")

    NUM_SCENES_DONE = 0
    for split, data in dataset.items():
        for i, scene in enumerate(tqdm(data)):
            write_object_affordances_for_scene(
                path + f"/object_data/{split}/{i}", scene, receptacle_mappings, pickupable_mappings, pickupables_blacklist=blacklist, unique_receptacles=unique_receptacles
            )
            if num_scenes_to_do is not None and NUM_SCENES_DONE >= num_scenes_to_do:
                return
            NUM_SCENES_DONE += 1


if __name__ == "__main__":
    curdir = "/".join(__file__.split("/")[:-1])
    # write_object_affordances_for_procthor(curdir, num_scenes_to_do=10)
    seed_everything(12345)
    write_object_affordances_for_procthor(curdir, unique_receptacles=False, num_scenes_to_do=10)
    ObjectCollection.create(
        curdir + "/object_data/test/0",
        jax.random.PRNGKey(0),
        limit_num_pickupables=20,
        limit_num_receptacles_per_pickupable=2,
        max_pickupable_per_receptacle=3,  # Set to -1 if no limit
        oob_receptacle_prob=0.1,
        bias_sampling=False
    )



import copy
import shutil
#os.environ["JAX_PLATFORM_NAME"] = "cpu"
import jax
#jax.config.update('jax_platform_name', "cpu")

import functools

import numpy as np

from semistaticsim.datawrangling.sssd import thin2wide
from semistaticsim.groundtruth import schedule_generation
from semistaticsim.groundtruth.objects import ObjectCollection
from semistaticsim.groundtruth.procthor_object import ProcThorObject
from semistaticsim.groundtruth.utils import seed_everything
import hydra

import json
import os
from tqdm import tqdm, trange

from flax import struct
import jax.numpy as jnp

from semistaticsim.rendering.floor import split_key
from semistaticsim.groundtruth.schedule import Schedule, TIME_SCALES_MAPPING
from semistaticsim.groundtruth.spoof_vmap import spoof_vmap
import tempfile
import polars as pl
from omegaconf import OmegaConf


OmegaConf.register_new_resolver("eval", eval)

def to_serializable(obj):
    """Recursively convert JAX or NumPy arrays to Python types."""
    if isinstance(obj, (jnp.ndarray, np.ndarray)):
        return obj.item()
    elif hasattr(obj, "item") and callable(obj.item):
        # scalar array
        try:
            return obj.item()
        except Exception:
            return float(obj)
    elif isinstance(obj, dict):
        return {k: to_serializable(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [to_serializable(v) for v in obj]
    else:
        return obj

@struct.dataclass
class Simulator:
    object_collection: ObjectCollection
    schedule: Schedule # one per pickupable
    current_positions: jnp.ndarray
    current_rotations: jnp.ndarray

    @property
    def num_pickupables(self) -> int:
        return self.object_collection.pickupables.position.shape[0]

    @property
    def num_receptacles(self) -> int:
        return self.object_collection.receptacles.position.shape[0]

    @property
    def pickupables(self):
        return self.object_collection.pickupables

    @property
    def receptacles(self):
        return self.object_collection.receptacles

    @property
    def receptacle_names(self):
        return [self.object_collection.receptacle_index_to_id(i) for i in range(self.num_receptacles)]

    @property
    def pickupable_names(self):
        return [self.object_collection.pickupable_index_to_id(i) for i in range(self.num_pickupables)]

    @property
    def current_assignment(self):
        return self.schedule.current_location

    #@property
    #@jax.jit
    def get_current_receptacles(self):#:, current_assignment=None):
        #if current_assignment is None:
        #    current_assignment = self.current_assignment
        current_assignment = self.current_assignment

        ret = jax.vmap(self.object_collection.receptacles.take)(current_assignment)
        return ret

    #@property
    #@jax.jit
    def get_receptacle_to_pickupable_mask(self):
        # which receptacle contains which pickupable?
        def collect(current_assignment, one_receptacle):
            return current_assignment == one_receptacle # mask of pickupables
        ret = jax.vmap(functools.partial(collect, self.current_assignment))(jnp.arange(self.num_receptacles))
        # ret.sum(axis=0) should be equal to np.ones(self.num_pickupables)
        return ret

    #def sample_positions(self, key):
    #    key, rngs = split_key(key, self.num_pickupables)
    #    ret = jax.vmap(ProcThorObject. lambda x,k: x.sample_from_surface(k))(self.current_receptacles, rngs)
    #    return ret.squeeze()

    #@jax.jit
    def get_positions_for_receptacle(self, receptacle_id):
        def thunk(r_id, pos, assignment):
            mask = r_id == assignment
            return jax.lax.cond(mask, lambda:pos, lambda:jnp.ones(3)*-jnp.inf)
        bound = functools.partial(thunk, receptacle_id)
        return jax.vmap(bound)(self.current_positions, self.current_assignment)

    #@property
    #@jax.jit
    def get_positions_mat(self):
        ret = jax.vmap(self.get_positions_for_receptacle)(jnp.arange(self.num_receptacles))
        return ret

    def sample_y_axis_rotations(self, key):
        return jax.random.uniform(key, minval=0, maxval=2*jnp.pi, shape=(self.num_pickupables,))

    #@jax.jit
    def sample_position_potential_field(self, key, old_self, spoof_assignment_changes=False):
        # the spoof argument is only useful when creating the simulator; we spoof a change to force a resample of all positions

        @functools.partial(jax.jit, static_argnums=0)
        def handle_one_receptacle(spoof_assignment_changes: bool, key, receptacle, old_this_r_to_all_p, new_this_r_to_all_p, old_positions_for_this_r):
            assignment_change_mask = old_this_r_to_all_p != new_this_r_to_all_p if not spoof_assignment_changes else jnp.ones_like(new_this_r_to_all_p)
            def get_static_positions_for_this_r(p,assignment_change_mask):
                # if no change happened, keep old position
                return jax.lax.cond(assignment_change_mask, lambda:jnp.ones(3) * -jnp.inf, lambda:p)
            static_positions_for_this_r = jax.vmap(get_static_positions_for_this_r)(old_positions_for_this_r, assignment_change_mask)

            new_positions = static_positions_for_this_r
            for i in jnp.arange(self.num_pickupables):
                def noop(k, potential_field, static_position_for_this_p):
                    return static_position_for_this_p

                def sample(k, potential_field, static_position_for_this_p):
                    potential_field = receptacle.make_potential_field(potential_field)
                    return receptacle.sample_from_surface_with_potential_field(k,potential_field).squeeze()

                key, rng = jax.random.split(key)
                newpos = jax.lax.cond(
                    jnp.logical_and(new_this_r_to_all_p.take(i), assignment_change_mask.take(i)), # must be assigned to this receptacle AND have just changed (meaning that we dont currently have the obj inside current_positions
                    sample,
                    noop,
                    rng, new_positions, new_positions.take(i, axis=0)
                )
                new_positions = new_positions.at[i].set(newpos)
            return new_positions


        # assert (jnp.nan_to_num(positions_mat, neginf=0).sum(axis=0) == self.current_positions).all()

        if old_self.current_positions is None:
            old_self = old_self.replace(current_positions=jnp.ones((self.num_pickupables , 3)) * -jnp.inf)
        if self.current_positions is None:
            self = self.replace(current_positions=jnp.ones((self.num_pickupables , 3)) * -jnp.inf)

        #handle_one_receptacle(key, self.receptacles.take(5), old_self.receptacle_to_pickupable_mask()[5],
        #                      self.receptacle_to_pickupable_mask()[5], old_self.get_positions_mat()[5])

        key, rngs = split_key(key, self.num_receptacles)
        new_positions = jax.vmap(functools.partial(handle_one_receptacle, spoof_assignment_changes))(rngs, self.receptacles, old_self.get_receptacle_to_pickupable_mask(), self.get_receptacle_to_pickupable_mask(), self.get_positions_mat())

        # noqa this is useful for non-jit debugging
        #gathered = []
        #for tm in new_positions:
        #    for i, p in enumerate(tm):
        #        if not (p == -jnp.inf).any():
        #            gathered.append(i)
        #assert len(set(gathered)) == self.num_pickupables

        #return handle_one_receptacle(key, self.receptacles.take(5), old_self.receptacle_to_pickupable_mask[5], self.receptacle_to_pickupable_mask[5], old_self.get_positions_for_receptacle(5))
        return jnp.nan_to_num(new_positions, neginf=0).sum(axis=0)


    @classmethod
    def create(cls, key, object_collection: ObjectCollection, schedule: Schedule):
        schedule = jax.tree.map(jnp.squeeze, schedule)
        self = cls(object_collection, schedule, current_positions=None, current_rotations=None)

        key, rng = jax.random.split(key)
        initial_positions = self.sample_position_potential_field(key, self, spoof_assignment_changes=True)
        self = self.replace(current_positions=initial_positions, current_rotations=self.sample_y_axis_rotations(key))
        return self

    #@jax.jit
    def step(self, key):
        key, rngs = split_key(key, self.num_pickupables)
        new_schedules = jax.vmap(lambda s,k: s.tick(k))(self.schedule, rngs)
        new_schedules = jax.tree.map(jnp.squeeze, new_schedules)
        new_self = self.replace(schedule=new_schedules)
        new_assignment = new_self.current_assignment

        assignment_changed_mask = jnp.astype(new_assignment != self.current_assignment, float)
        key, rng = jax.random.split(key)
        # if assignment changed, the object moved, and we need to resample a y axis rotation
        new_rotations = self.current_rotations * (1-assignment_changed_mask) + self.sample_y_axis_rotations(rng) * assignment_changed_mask

        key, rng = jax.random.split(key)
        maybe_new_positions = new_self.sample_position_potential_field(rng, self)
        #new_positions = jax.vmap(lambda old, new, mask: old * (1-mask) + new * mask)(self.current_positions, maybe_new_positions, assignment_changed_mask) # not actually necessary because this is handled in the sample_potential_field function; left for readability only

        new_self = new_self.replace(current_positions = maybe_new_positions, current_rotations = new_rotations)
        return new_self

    #@jax.jit
    def diff(self, old_self):
        assignment_change = self.current_assignment != old_self.current_assignment
        position_before_and_after = jnp.concatenate((old_self.current_positions[None], self.current_positions[None]))
        # could also get the duratin left from the schedules etc if necessary
        return assignment_change, position_before_and_after


    def make_scan_thunk(self, scansize, scanunroll=10):
        #@jax.jit
        def thunk(self, key):
            old_self = self
            self = self.step(key)
            return self, (self.current_assignment, self.current_positions, self.current_rotations)#, *old_self.diff(self))

        def scan_it(key, self):
            self, data = jax.lax.scan(thunk, self, split_key(key, scansize)[-1], unroll=min(scanunroll, scansize))
            ret = {k:d for k,d in zip(["assignment", "position", "rotation"], data)}
            return self, ret

        return jax.jit(scan_it)

        #return jax.jit(lambda key, simulator: jax.lax.scan(thunk, simulator, split_key(key, scansize)[-1], unroll=min(scanunroll, scansize)))
        #return thunk


@hydra.main(version_base=None, config_path="config", config_name="config")
def _main(cfg):
    bare_cfg = copy.copy(cfg)

    schedule_generation_func = functools.partial(schedule_generation.full_pattern, **cfg.schedule)

    seed_everything(cfg.seed)
    key = jax.random.PRNGKey(cfg.seed)
    key, rng = jax.random.split(key)
    object_collection: ObjectCollection = hydra.utils.call(cfg.objects, key=rng)

    schedules = []
    for i, object in enumerate(tqdm(object_collection.pickupable_to_receptacles.keys())):
        location_indices = object_collection.valid_receptacle_indices_for_pickupable_index(i)
        location_weights = object_collection.mat_pickupable_to_receptacles[i,location_indices]
        num_pad_locations = object_collection.mat_pickupable_to_receptacles.shape[1] - len(location_indices)
        location_indices = np.array(location_indices).tolist()

        key, rng = jax.random.split(key)
        schedule_pattern =  schedule_generation_func(
                key=rng,
                locations=location_indices,
                locations_weights=location_weights,
                num_pad_locations=num_pad_locations,
                dt=cfg.dt
            )

        location_names = [object_collection.receptacle_index_to_id(i) for i in location_indices]

        key, rng = jax.random.split(key)
        schedule = Schedule.create(rng, schedule_pattern, scales=cfg.schedule.scales, locations= location_indices + [-1] * num_pad_locations, dt=cfg.dt)
        schedules.append(schedule)

    schedules = spoof_vmap(schedules)
    simulator = Simulator.create(key=jax.random.PRNGKey(0), object_collection=object_collection, schedule=schedules)

    #simulator = simulator.step(key=jax.random.PRNGKey(0))

    timespan = cfg.timespan_to_collect
    num, unit = timespan.split()
    num = int(num)
    unit = TIME_SCALES_MAPPING[unit]
    timespan = num * unit * (1 / simulator.schedule.dt[0])
    num_of_scans_to_do = timespan // cfg.jax_scan_size
    if num_of_scans_to_do * simulator.schedule.dt[0] < timespan:
        num_of_scans_to_do += 1

    print("Number of time to get:", cfg.timespan_to_collect)
    print("Number of time in terms of units:", num * unit)
    del num
    del unit
    print("Current dt:", simulator.schedule.dt[0])
    print("Number of dt we need to fill timespan to collect:", timespan)
    print(f"Number of scans to do with scansize {cfg.jax_scan_size}:", num_of_scans_to_do)

    #with jax.disable_jit():
    thunk = simulator.make_scan_thunk(cfg.jax_scan_size, cfg.jax_scan_unroll)

    # create a temporary directory under /tmp
    tmpdir = tempfile.mkdtemp(dir="/tmp")

    with open(os.path.join(tmpdir, "pickupable_to_receptacle.json"), 'w') as f:
        pickupable_to_receptacles = {}
        for pickupable_id, row in enumerate(simulator.object_collection.mat_pickupable_to_receptacles):
            pickupable_to_receptacles[simulator.pickupable_names[pickupable_id]] = [simulator.receptacle_names[i] for i in np.array(jnp.nonzero(row > 0)).tolist()[0]]
        json.dump(pickupable_to_receptacles, f, indent=4)

    # Save receptacles' pose
    receptacles_oobb = {}
    for id, receptacle in enumerate(simulator.receptacle_names):
        obj_metadata = simulator.object_collection.get_receptacle_by_index(id)
        oobb = obj_metadata.get_oobb()
        receptacles_oobb[receptacle] = to_serializable(oobb)
        
    with open(os.path.join(tmpdir, "receptacles_oobb.json"), 'w') as f:
        json.dump(receptacles_oobb, f, indent=4)

    with open(os.path.join(tmpdir, "receptacle_names.json"), 'w') as f:
        json.dump(simulator.receptacle_names, f, indent=4)

    with open(os.path.join(tmpdir, "pickupable_names.json"), 'w') as f:
        json.dump(simulator.pickupable_names, f, indent=4)

    for step in trange(int(num_of_scans_to_do)):
        key, rng = jax.random.split(key)
        simulator, data = thunk(key, simulator)
        timestamps = (jnp.arange(cfg.jax_scan_size) + step * cfg.jax_scan_size) * simulator.schedule.dt[0]
        data['timestamp'] = timestamps[:, jnp.newaxis]
        data["assignment"] = thin2wide(data["assignment"], simulator.receptacle_names)

        if cfg.output_bboxes:
            aabb, oobb = jax.vmap(ProcThorObject.resolve_bboxes, in_axes=(0,1,1))(simulator.object_collection.pickupables, data["position"], data["rotation"])
            data["aabb_center"] = aabb["center"].transpose(1,0,2)
            data["aabb_cornerPoints"] = aabb["cornerPoints"].transpose(1,0,2,3)
            data["aabb_size"] = aabb["center"].transpose(1,0,2) # jnp.concatenate((aabb["size"]["x"][None], aabb["size"]["y"][None], aabb["size"]["z"][None])).transpose(2,1,0)
            data["oobb_cornerPoints"] = oobb["cornerPoints"].transpose(1,0,2,3)

        data = {k:np.array(v) for k,v in data.items()}

        data = pl.DataFrame(data)

        out_path = os.path.join(tmpdir, f"scan_{step}.parquet")
        data.write_parquet(out_path)

    print(f"Dumped data to <{tmpdir}>, now moving it to <{cfg.target_dir}>")
    # Delete existing results
    if os.path.exists(cfg.target_dir):
        shutil.rmtree(cfg.target_dir)
    os.makedirs(cfg.target_dir, exist_ok=True)

    for filename in os.listdir(tmpdir):
        src_path = os.path.join(tmpdir, filename)
        dst_path = os.path.join(cfg.target_dir, filename)
        shutil.move(src_path, dst_path)
    with open(cfg.target_dir + "/config.yaml", "w") as f:
        OmegaConf.save(cfg, f, resolve=True)
    #shutil.move(f"{tmpdir}/", cfg.target_dir)

        # todo log data as parquet file


def main(version_base=None, config_path="config", config_name="config"):
    try:
        _main(version_base=None, config_path="config", config_name="config")
    except Exception as e:
        print(f"An error occurred: {e}")
        print("Attempting to fall back to default config")
        return _main(
            version_base=None,
            config_path=os.path.join(os.path.dirname(__file__), "config"),
            config_name="config",
        )

if __name__ == '__main__':
    main()



import copy


def get_robot_inventory(controller, agent_id):
    inventory = controller.last_event.events[agent_id].metadata["inventoryObjects"]
    inventory = [o["objectId"] for o in inventory]

    return inventory

def get_object_from_controller(controller, object_id):
    # you should _verify_object_exists first
    for obj in get_object_list_from_controller(controller):
        if obj["objectId"] == object_id:
            return obj
    return None

def get_object_aabb_from_controller(controller, object_id):
    return get_object_from_controller(controller, object_id)["axisAlignedBoundingBox"]

def get_object_size_from_controller(controller, object_id):
    aabb = get_object_aabb_from_controller(controller, object_id)['size']
    return (aabb['x'], aabb['y'], aabb['z'])

import numpy as np
def compute_aabb_distance(controller, pickupable_name, receptacle_name):
    OBJECT_METADATA = {o["objectId"]: o for o in controller.last_event.metadata["objects"]}

    aabb_p = OBJECT_METADATA[pickupable_name]["axisAlignedBoundingBox"]
    aabb_r = OBJECT_METADATA[receptacle_name]["axisAlignedBoundingBox"]

    # Extract centers and half-sizes
    c_p = np.array([aabb_p["center"]["x"], aabb_p["center"]["y"], aabb_p["center"]["z"]])
    s_p = np.array([aabb_p["size"]["x"], aabb_p["size"]["y"], aabb_p["size"]["z"]]) / 2.0

    c_r = np.array([aabb_r["center"]["x"], aabb_r["center"]["y"], aabb_r["center"]["z"]])
    s_r = np.array([aabb_r["size"]["x"], aabb_r["size"]["y"], aabb_r["size"]["z"]]) / 2.0

    # Compute per-axis distances
    delta = np.abs(c_p - c_r) - (s_p + s_r)
    delta = np.maximum(delta, 0.0)  # No negative distances (overlaps → 0)

    # Euclidean distance between the boxes
    distance = np.linalg.norm(delta)
    return distance

def get_object_position_from_controller(controller, object_id):
    aabb = get_object_aabb_from_controller(controller, object_id)
    pos = aabb["center"]
    return (pos['x'], pos['y'], pos['z'])

THUNK_DIR = 1
def thunk_fix_robot_pos(controller):
    #return
    global THUNK_DIR

    # print("Thunk Fix Robot Pos Called")
    act = {
        0: "MoveAhead",
        1: "MoveBack",
        2: "MoveLeft",
        3: "MoveRight",
    }
    #time.sleep(0.001)
    controller.step(
        dict(action=act[THUNK_DIR], moveMagnitude=0.000001,
             agentId=0))  # for some reason, non-move actions tend to alter the robot position in unexpected ways. To fix this, we spoof a fake movement to reset it
    THUNK_DIR += 1
    THUNK_DIR %= 4

def get_robot_position_from_controller(controller, robot_id):
    thunk_fix_robot_pos(controller)
    #metadata = controller.last_event.events[robot_id].metadata
    metadata = controller.last_event.metadata
    pos = [metadata["agent"]["position"]["x"],
        metadata["agent"]["position"]["y"],
        metadata["agent"]["position"]["z"]]
    return pos

def get_object_list_from_controller(controller):
    objects = controller.last_event.metadata["objects"]
    objects = copy.deepcopy(objects)

    held_objects = {} # id: robot
    for i, robot_event in enumerate(controller.last_event.events):
        robot_metadata = robot_event.metadata

        inventory = robot_metadata["inventoryObjects"]

        thunk_fix_robot_pos(controller)
        robotpos = robot_metadata["agent"]['position']
        SIZE = {'x': 0.4, 'y': 1.0, 'z': 0.4}
        robotpos["x"] = robotpos["x"] + SIZE["x"] / 2
        robotpos["y"] = robotpos["y"] - SIZE["y"] / 2
        robotpos["z"] = robotpos["z"] + SIZE["z"] / 2

        robot_dict = {
            "assetId": f"", # makes robot not considered as runtime object
            "objectId": f"robot{i+1}",
            "id": i,
            "position": robotpos,
            "rotation": robot_metadata["agent"]['rotation'],
            "size": SIZE,
            "inventory": inventory,
            "ISROBOT": True
        }
        for obj in inventory:
            held_objects[obj["objectId"]] = f"robot{i+1}"

        objects.append(robot_dict)

    for obj_dict in objects:
        if obj_dict["objectId"] in held_objects:
            obj_dict["heldBy"] = held_objects[obj_dict["objectId"]]
        else:
            obj_dict["heldBy"] = None

    return objects



def recursive_map(data, func):
    if isinstance(data, dict):
        return {k: recursive_map(v, func) for k, v in data.items()}
    elif isinstance(data, (list, tuple, set)):
        return type(data)(recursive_map(v, func) for v in data)
    else:
        return func(data)


import networkx as nx

def get_scc(mat):
    # Build a directed graph from the transition matrix
    G = nx.from_numpy_array(mat, create_using=nx.DiGraph)

    # Find all strongly connected components
    sccs = list(nx.strongly_connected_components(G))

    # Pick the largest one
    largest_scc = max(sccs, key=len)
    return list(largest_scc)


from typing import List, Tuple

from functools import partial

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import seaborn as sns

from datetime import datetime, timedelta
import numpy as np
import os
import datetime
from pathlib import Path
os.environ["JAX_PLATFORM_NAME"] = "cpu"
import jax
jax.config.update('jax_platform_name', "cpu")

from semistaticsim.datawrangling.sssd import load_sssd
from semistaticsim.groundtruth.utils import format_receptacle

# plt.style.use(["science", "latex-sans", "sans"])
plt.style.use(["science", "sans", "no-latex"])

color_palette = [(74, 35, 119), (140, 197, 227), (245, 95, 116), (89, 168, 156), (240, 197, 113)]
color_palette = [(r / 255, g / 255, b / 255) for r, g, b in color_palette]
color_palette_str = [
    "purple",
    "light_blue",
    "pink",
    "teal",
    "mustard",
]
color_palette_from_str = {k: v for k, v in zip(color_palette_str, color_palette)}

N_COLORS = 10
grey_palette = sns.color_palette("Greys", n_colors=N_COLORS+2)

# This is a grey palette: 0 means white, 100 means black
# The line below helps indexing "10" as '10% black', while "80" means '80% black
grey_palette_from_str = {f"{p}0": v for p, v in enumerate(grey_palette)}

# The stuff below is just to facilitate semantics
grey_palette_from_str["black"] = grey_palette_from_str["100"]
grey_palette_from_str["white"] = grey_palette_from_str["00"]

plt.rcParams["font.size"] = 8

def hours_to_datetimed(hours: np.ndarray, start_day: str = "monday") -> np.ndarray:
    """
    Convert minutes (from Monday 00:00) into datetime objects.
    """
    days_of_week = ["monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"]
    start_idx = days_of_week.index(start_day.lower())
    start_date = datetime.datetime(2023, 1, 2 + start_idx)  # pick arbitrary Monday
    return np.array([start_date + timedelta(hours=int(m)) for m in hours])

def save_fig(
    fig: matplotlib.figure.Figure,
    fig_name: str,
    fig_dir: str,
    fig_fmt: str,
    fig_size: Tuple[float, float] = [6.4, 4],
    save: bool = True,
    dpi: int = 300,
    transparent_png=True,
):
    """
    Code adapted from https://zhauniarovich.com/post/2022/2022-09-matplotlib-graphs-in-research-papers/
    This procedure stores the generated matplotlib figure to the specified
    directory with the specified name and format.

    Parameters
    ----------
    fig : [type]
        Matplotlib figure instance
    fig_name : str
        File name where the figure is saved
    fig_dir : str
        Path to the directory where the figure is saved
    fig_fmt : str
        Format of the figure, the format should be supported by matplotlib
        (additional logic only for pdf and png formats)
    fig_size : Tuple[float, float]
        Size of the figure in inches, by default [6.4, 4]
    save : bool, optional
        If the figure should be saved, by default True. Set it to False if you
        do not want to override already produced figures.
    dpi : int, optional
        Dots per inch - the density for rasterized format (png), by default 300
    transparent_png : bool, optional
        If the background should be transparent for png, by default True
    """
    if not save:
        return

    fig.set_size_inches(fig_size, forward=False)
    fig_fmt = fig_fmt.lower()
    fig_dir = os.path.join(fig_dir, fig_fmt)
    if not os.path.exists(fig_dir):
        os.makedirs(fig_dir)
    pth = os.path.join(fig_dir, "{}.{}".format(fig_name, fig_fmt.lower()))
    if fig_fmt == "pdf":
        metadata = {"Creator": "", "Producer": "", "CreationDate": None}
        fig.savefig(pth, bbox_inches="tight", metadata=metadata)
    elif fig_fmt == "png":
        alpha = 0 if transparent_png else 1
        axes = fig.get_axes()
        fig.patch.set_alpha(alpha)
        for ax in axes:
            ax.patch.set_alpha(alpha)
        fig.savefig(
            pth,
            bbox_inches="tight",
            dpi=dpi,
        )
    else:
        try:
            fig.savefig(pth, bbox_inches="tight")
        except Exception as e:
            print("Cannot save figure: {}".format(e))





class Plotter:
    def __init__(self, exp_name: str, viz: bool = True, res_path: str = None, fig_fmt="pdf") -> None:
        self.viz = viz

        # Create a directory to store the results if viz is False
        script_directory = os.path.abspath(os.curdir)
        self.res_path = res_path
        if self.res_path is None:
            self.res_path = os.path.join(script_directory, "assets", datetime.datetime.now().isoformat(), exp_name)
        print(f"Saving results to {self.res_path}")
        if not os.path.exists(self.res_path):
            os.makedirs(self.res_path)
        # Conversion utils
        self.mm = 1 / 25.4  # conversion from millimeters to inches
        self.cm = 1 / 2.54  # conversion from centimeters to inches
        self.pt = 1 / 72.27  # conversion from points to inches

        # The values below are obtained from LaTeX, using \the\columnwidth and \the\textwidth
        self.PAPER_COLUMN = 252.0  # points
        self.PAPER_TEXTWIDTH = 516.0  # points
        # Savefig partial
        self.savefig = partial(save_fig, fig_dir=self.res_path, fig_fmt=fig_fmt, transparent_png=False)

    def plot_observations(
        self,
        sampling_times: np.ndarray,
        assignments: np.ndarray,
        receptacles_name: List[str],
        pickupable_name: str,
        location: str = "lower left",
        f_name: str = "ground_truth",
    ) -> None:
        """
        Plot binary observations against the true state over time.

        Args:
            sampling_times (List[float]): Time points for the true state samples.
            true_state (List[float]): True state values at sampling times.
            t (List[float]): Time points for the observations.
            obs (List[int]): Binary observations at times `t`.
            location (str, optional): Legend location in the plot. Default is "lower left".
            figure_size (Optional[Tuple[int, int]], optional): Figure size in inches. Defaults to `self.figure_size`.
            f_name (str, optional): File name for saving the plot. Default is "observations".

        Returns:
            None
        """
        # Obtain max mumber of unique receptacles across all pickupables (columns)
        n_receptacles = len(np.unique(assignments))

        im_width, im_height = self.PAPER_TEXTWIDTH * self.pt, 4 * self.cm * n_receptacles 
        figure_size = (im_width, im_height)

        # Create figure
        fig, axes = plt.subplots(n_receptacles, 1, figsize=figure_size, sharex=True, dpi=300, constrained_layout=True)
        if n_receptacles == 1:
            axes = [axes]

        dates = hours_to_datetimed(sampling_times, "monday")

        for _ in range(n_receptacles):
            receptacles_id = np.unique(assignments).astype(int).tolist()
            for i, receptacle_id in enumerate(receptacles_id):
                mask = assignments == receptacle_id
                states = mask.astype(int)
                # format receptacle name
                r_name = receptacles_name[receptacle_id]
                r_name = format_receptacle(r_name)
                # Plot state
                axes[i].plot(dates, states, color=color_palette_from_str["purple"], ls="-", linewidth=0.5)
                axes[i].set_title(f"{pickupable_name} @ {r_name}")
                print(f"{pickupable_name} @ {r_name}")
                # Set y limits to 1 = present and 0 = absent
                axes[i].set_ylim(-0.1, 1.1)
                axes[i].set_yticks([0, 1])
                axes[i].set_yticklabels(["Absent", "Present"])
                # axes[p_id, i].xaxis.set_major_locator(mdates.MonthLocator())             # one tick per month
                # axes[p_id, i].xaxis.set_major_formatter(mdates.DateFormatter("%b"))      # Jan, Feb, Mar, ...
                axes[i].xaxis.set_major_locator(mdates.DayLocator())  # ticks per day
                axes[i].xaxis.set_major_formatter(mdates.DateFormatter("%a"))  # Mon, Tue, ...
                # Make x label bigger
                axes[i].set_xlabel("Weekday")

        if self.viz:
            self.savefig(fig, fig_name=f_name, save=False, fig_size=figure_size)
            plt.show()
        else:
            self.savefig(fig, fig_name=f_name, save=True, fig_size=figure_size)

        plt.close()


def plot_ground_truth():
    # Load a config file
    # Use path library to parse the path
    folder = Path("generated_data/semistaticsim/test/0/0/groundtruth")
    sssd_data = load_sssd("generated_data/semistaticsim/test/0/0/groundtruth")
    assignments: List[np.ndarray] = []
    timestamps: List[np.ndarray] = []
    for data in sssd_data.get_generator_of_selves():
        assignments.append(data._assignment)
        timestamps.append(data._timestamp)

    assignments = np.concatenate(assignments, axis=0)
    # Make assignments thing format and not one-hot
    assignments = np.argmax(assignments, axis=2)
    timestamps = np.concatenate(timestamps, axis=0)
    pickupables = sssd_data.pickupable_names
    receptacles = sssd_data.receptacle_names

    # Hyperparameters
    plotter = Plotter(exp_name="ground_truth", viz=False, res_path=folder, fig_fmt="png")

    # Plot first pickupable for now
    for pickupable_id in range(len(pickupables)):
        pickupable_name = pickupables[pickupable_id].split("|")[0]
        receptacles_assignments = assignments[:, pickupable_id]
        plotter.plot_observations(
            sampling_times=timestamps,
            assignments=receptacles_assignments,
            receptacles_name=receptacles,
            pickupable_name=pickupable_name,
            location="upper right",
            f_name=f"pickupable_{pickupable_id}_ground_truth",
        )
        print(f"Plotted {pickupable_name}")


if __name__ == "__main__":
    plot_ground_truth()



AFFORDANCE_NAMES = [
    "receptacle",
    "toggleable",
    "breakable",
    "canFillWithLiquid",
    "dirtyable",
    "canBeUsedUp",
    "cookable",
    "sliceable",
    "openable",
    "pickupable",
    "moveable",
]
